"""Miscellaneous utility methods for collections."""

import typing as _t
from collections.abc import Iterable as _Iterable

ArgType = _t.TypeVar("ArgType")
"""ArgType generic type."""


def as_list(arg: ArgType | _Iterable[ArgType] | None = None, excl_types: tuple[type[_t.Any]] = (str,)) -> list[ArgType]:
    """Create a list or list-wrapping of `arg`.

    Args:
        arg: An object to name.
        excl_types: Iterable types that should be treated as single elements, such as strings.

    Returns:
        A list representation of `arg`.

    Notes:
        For all zero-length arguments, i.e. ``len(arg) == 0``, an empty list is returned.

    """
    # https://github.com/python/mypy/issues/10835
    if isinstance(arg, _Iterable) and not isinstance(arg, excl_types):
        return list(arg)
    else:
        return [arg]  # type: ignore[list-item]



"""Jupyter kernel installation for virtual environments.

For details, see the :doc:`CLI documentation </documentation/cli/notebooks/kernel>` page.
"""

import click

from rics import click as _rc
from rics import jupyter as _rj


@click.command("kernel")
@click.pass_context
@click.option(
    "--user/--system",
    is_flag=True,
    default=None,
    help=_rj._kernel_helper.DOCSTRINGS["user"],
)
@click.option(
    "--name",
    type=str,
    help=_rj._kernel_helper.DOCSTRINGS["display_name"],
    show_default="automatic",
)
@click.option(
    "--profile",
    type=str,
    help=_rj._kernel_helper.DOCSTRINGS["profile"],
)
@click.option(
    "--prefix",
    type=str,
    help=_rj._kernel_helper.DOCSTRINGS["prefix"] + " Mutually exclusive with --user."
    "\n\n\b\nThe kernel JSON will be located at:\n   `<prefix>/share/jupyter/kernels/rics.kernel.<slug>/kernel.json`.",
)
@click.option(
    "--env",
    "-e",
    type=(str, str),
    multiple=True,
    help=_rj._kernel_helper.DOCSTRINGS["env"]
    + "\n\n\b\nGiven as [<Key> <Value>]-pairs. Example:\n    `<cmd> -e OMP_NUM_THREADS 32 -e MY_VAR my_value`",
)
@click.option(
    "--variant",
    help="Install a variant of the kernel, e.g. with extra environment variables.",
)
@click.option(
    "--add-package",
    "-p",
    type=str,
    default=(
        "jupyterlab-execute-time",
        "jupyterlab-git",
        "jupyterlab-code-formatter",
        "ipywidgets",
        "black",
        "isort",
    ),
    multiple=True,
    show_default=True,
    help="Packages to install in the kernel. Pass a single character to disable extra packages, e.g. `-p0`.",
)
def main(
    cxt: click.Context,
    *,
    user: bool | None,
    name: str | None,
    prefix: str | None,
    profile: str | None,
    env: tuple[tuple[str, str], ...],
    variant: str | None,
    add_package: tuple[str, ...],
) -> None:
    """Install Jupyter kernel spec for a virtual environment."""
    extra_packages = [] if len(add_package) == 1 and len(add_package[0]) == 1 else add_package
    helper = _rj.KernelHelper(
        extra_packages=extra_packages,
        variant=variant,
        callback=_Callback(cxt),
    )

    display_name = helper.resolve_display_name(name)
    click.secho(
        f"Installing KernelSpec({display_name=}, name={helper.resolve_kernel_name()!r}, executable='{helper.venv.executable}').",
        fg="green",
    )

    path = helper.install(
        user=user,
        display_name=display_name,
        prefix=prefix,
        profile=profile,
        env=dict(env),
        frozen_modules=True,
    )
    click.secho(f"Kernel spec installed: '{path}'", fg="green")


class _Callback:
    def __init__(self, cxt: click.Context) -> None:
        self.cxt = cxt

    def write_user_command(self, kernel_spec: _rj.KernelSpec) -> None:
        installer_metadata = kernel_spec["metadata"][__package__]["installer"]
        installer_metadata["user_command"] = _rc.get_user_command(self.cxt)

    __call__ = write_user_command
    __name__: str = __call__.__name__


if __name__ == "__main__":
    # TODO(alternative): Creates a new kernel spec for each project. Something like https://pypi.org/project/pyproject-local-kernel/
    #  could be an interesting alternative. But it's more opaque, and seems outdated (looks for tool.poetry.name).
    main()



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click

if TYPE_CHECKING:
    import pandas

_DATETIME_DTYPE = "datetime64[ns, UTC]"
_PUSH = "Last push"
_SORT_COLUMNS = {
    "push": (_PUSH, False),
    "stars": ("Stars", False),
    "created": ("Created", True),
    "name": ("Repo name", True),
}


@click.command("repos")
@click.option(
    "--user",
    "-u",
    help="User or organization on GitHub.",
    default=["rsundqvist"],
    multiple=True,
    show_default=True,
)
@click.option(
    "--max-days",
    "-d",
    type=click.IntRange(min=0),
    help="Max days since the last push.",
)
@click.option(
    "--sort",
    type=click.Choice(sorted(_SORT_COLUMNS), case_sensitive=False),
    help="Column to sort on.",
    default="push",
    show_default=True,
)
@click.option(
    "--max-width",
    type=int,
    help="Maximum width for the 'Description' column. Zero=no limit.",
)
def main(
    max_days: int | None,
    user: tuple[str, ...],
    sort: str,
    max_width: int | None,
) -> None:  # pragma: no coverage
    """Get licenced GitHub repositories with releases."""
    import functools
    import logging
    import sys
    from concurrent.futures import ThreadPoolExecutor
    from time import perf_counter

    import click
    import pandas as pd
    import requests

    from .strings import format_perf_counter

    now = pd.Timestamp.utcnow()
    logger = logging.getLogger(f"{__package__}.{main.name}")

    headers = _get_headers(logger)

    per_page = 100

    start = perf_counter()
    raw = []
    for i, u in enumerate(user, start=1):
        n = 0
        page = 1
        while True:
            response = requests.get(
                f"https://api.github.com/users/{u}/repos?{page=}&{per_page=}",
                headers=headers,
                timeout=5,
            )
            response.raise_for_status()
            user_repos = response.json()

            n_repos = len(user_repos)
            n += n_repos
            raw.extend(user_repos)

            if n_repos < per_page:
                args = u, page, n_repos, per_page
                logger.debug("Halt user=%r at page=%i; query returned %i repos < per_page=%i.", *args)
                break
            page += 1

        logger.debug("Found %i repo(s) for user %i/%i: '%s'.", n, i, len(user), u)

    logger.info("Found %i repo(s) for %i user(s): %s.", len(raw), len(user), ", ".join(map(repr, user)))

    func = functools.partial(_process_repo, logger=logger, headers=headers, now=now, max_days=max_days)
    with ThreadPoolExecutor(max_workers=8, thread_name_prefix=logger.name) as pool:
        repos = pool.map(func, raw, timeout=300)

    filtered_repos = [r for r in repos if r is not None]
    logger.info("Processed %i/%i matching repos in %s.", len(filtered_repos), len(raw), format_perf_counter(start))

    if not repos:
        click.secho("No repositories found.", fg="green")
        sys.exit(0)

    df = _to_dataframe(filtered_repos, sort)
    pretty = _to_string(df, max_width)
    click.secho(pretty, fg="green")


def _get_headers(logger: logging.Logger) -> dict[str, str] | None:
    if not (path := Path("~/.rics/github-token").expanduser()).is_file():
        logger.debug("Skipping GitHub auth; file not found: '%s'.", path)
        return None

    logger.debug("Found GitHub auth token: '%s'.", path)
    return {
        "Authorization": f"Bearer {path.read_text().strip()}",
        "X-GitHub-Api-Version": "2022-11-28",
    }


def _process_repo(
    repo: dict[str, Any],
    *,
    logger: logging.Logger,
    headers: dict[str, str] | None,
    now: "pandas.Timestamp",
    max_days: int | None,
) -> "dict[str, str | int | pandas.Timestamp] | None":
    import pandas as pd
    import requests

    if not repo.get("license"):
        logger.debug(f"Skip repo={repo['full_name']!r}: no licence.")
        return None

    url = repo["releases_url"].replace("{/id}", "/latest")
    response = requests.get(url, headers=headers, timeout=5)
    if response.status_code == 404:  # noqa: PLR2004
        logger.debug(f"Skip repo={repo['full_name']!r}: no releases.")
        return None

    release = response.json()
    pushed_at = pd.Timestamp(repo["pushed_at"])
    age = (now - pushed_at).days

    if max_days is not None and age > max_days:
        logger.debug(f"Skip repo={repo['full_name']!r}: {age=} > {max_days=}.")
        return None

    created_at = pd.Timestamp(repo["created_at"])
    return {
        "Repo name": repo["full_name"],
        "Release tag": release["tag_name"],
        _PUSH: pushed_at,
        # "Age [days]": age,
        "Stars": repo["stargazers_count"],
        "Issues": repo["open_issues"],
        "Description": repo["description"],
        "Link": repo["html_url"],
        "License": repo["license"]["spdx_id"],
        "Created": created_at,
        # "Created [days]": (now - created_at).days,
    }


def _to_dataframe(repos: list[dict[str, Any]], sort: str) -> "pandas.DataFrame":
    import pandas as pd

    df = pd.DataFrame.from_records(repos)
    assert (df.dtypes[[_PUSH, "Created"]] == _DATETIME_DTYPE).all()  # noqa: S101

    (sort, ascending) = _SORT_COLUMNS[sort]
    df = df.sort_values(sort, ascending=ascending)
    df = df.set_index("Repo name")

    return df


def _to_string(df: "pandas.DataFrame", max_width: int | None) -> str:
    if max_width is None:
        max_width = _max_width()
    if max_width > 0:
        df["Description"] = df["Description"].str[:max_width]

    pretty = df.to_string(
        max_colwidth=None,
        formatters={_PUSH: "{:%Y-%m-%dT%H:%M:%SZ}".format, "Created": "{:%Y-%m-%d}".format},
    )
    assert isinstance(pretty, str)  # noqa: S101
    return pretty


def _max_width() -> int:
    import os

    try:
        max_width = min(200, max(60, 40 - os.get_terminal_size()[0]))
    except OSError as e:
        from errno import ENOTTY

        if e.errno == ENOTTY:
            max_width = 80
        else:
            raise
    return max_width



"""Utility methods for logging tasks."""

import logging as _l
import typing as _t
from contextlib import contextmanager as _contextmanager

BASE_FORMAT = "[%(name)s:%(levelname)s] %(message)s"
FORMAT_MS: str = "%(asctime)s.%(msecs)03d " + BASE_FORMAT
"""Default logging format; ``<date-format>.378 [rics:DEBUG] I'm a debug message!``"""
FORMAT = FORMAT_MS  # Backwards compatibility
"""Alias of :attr:`FORMAT_MS`."""
FORMAT_SEC: str = "%(asctime)s " + BASE_FORMAT
"""Like :attr:`FORMAT_MS`, but without milliseconds; ``<date-format> [rics:DEBUG] I'm a debug message!``"""

DATE_FORMAT: str = "%Y-%m-%dT%H:%M:%S"
"""Default logging date format; ``2022-02-05T11:17:05<logging-format>``"""

LogLevel = int | str
"""Valid input types for :func:`convert_log_level`."""


def basic_config(
    *,
    format: str = FORMAT_MS,
    datefmt: str = DATE_FORMAT,
    level: LogLevel | None = "INFO",
    force: bool = True,
    **kwargs: _t.Any,
) -> None:
    """Do basic logging configuration with package defaults.

    Simple wrapper for the standard :py:func:`logging.basicConfig`-method, using my personal preferences for defaults.

    Args:
        format: Format string for emitted messages; see :attr:`FORMAT_MS`.
        datefmt: Format string for date/time; see :attr:`DATE_FORMAT`. If empty, remove the time components if
         ``format == FORMAT_MS`` (default) or ``FORMAT_SEC``.
        level: Log level for the root logger.
        force: If ``True``, override existing configuration if it exists.
        **kwargs: Keyword arguments for :py:func:`logging.basicConfig`.

    Keyword Args:
        _level: Additional log levels to set, replacing *double* underscores with dots to produce the logger names. For
            example, passing ``rics__performance_level=logging.INFO`` (or the dotted-name equivalent
            ``**{"rics.performance_level": "INFO"}``) will invoke
            :meth:`logging.getLogger("rics.performance").setLevel(logging.INFO) <logging.Logger.setLevel>`.

    Examples:
        Basic usage.

        >>> import logging
        >>> basic_config(level=logging.INFO, rics_level=logging.DEBUG)
        >>> logging.getLogger("rics").debug("I'm a debug message!")
        >>> logging.debug("I'm a debug message!")
        >>> logging.critical("I'm a critical message!")  # Doctest: +SKIP
        2022-02-05T11:17:05.378 [rics:DEBUG] I'm a debug message!
        2022-02-05T11:17:05.378 [root:CRITICAL] I'm a critical message!

        Removing time from the message template.

        >>> import sys
        >>> basic_config(datefmt="", stream=sys.stdout)
        >>> logging.info("No time!")
        [root:INFO] No time!

        Note that this only works when `format` is either :attr:`FORMAT_MS` or :attr:`FORMAT_SEC`.
    """
    if datefmt == "" and format in {FORMAT_MS, FORMAT_SEC}:
        format = BASE_FORMAT

    extra_levels, kwargs = _extract_extra_levels(**kwargs)
    _l.basicConfig(level=level, format=format, datefmt=datefmt, force=force, **kwargs)

    for name, extra_level in extra_levels.items():
        _l.getLogger(name).setLevel(extra_level)


@_contextmanager
def disable_temporarily(
    *loggers: str | _l.Logger | _l.LoggerAdapter,  # type: ignore[type-arg]
) -> _t.Any:
    """Temporarily disable logging.

    Args:
        *loggers: Loggers to disable.

    Yields:
        Nothing.

    Examples:
        Disable all logging temporarily.

        >>> import logging
        >>> logging.basicConfig()
        >>> with disable_temporarily(logging.root):
        ...     logging.info("This message is ignored.")

    """
    enabled_loggers: list[_l.Logger] = []
    for logger in loggers:
        while isinstance(logger, _l.LoggerAdapter):
            logger = logger.logger  # noqa: PLW2901
        if isinstance(logger, str):
            logger = _l.getLogger(logger)  # noqa: PLW2901

        if not logger.disabled:
            enabled_loggers.append(logger)

    try:
        for logger in enabled_loggers:
            logger.disabled = True
        yield
    finally:
        for logger in enabled_loggers:
            logger.disabled = False


class LogLevelError(Exception):
    """Raised by :func:`convert_log_level`."""

    def __init__(self, log_level: LogLevel, argument_name: str, extra_notes: _t.Collection[str] = ()) -> None:
        super().__init__(f"Unknown {argument_name}={log_level!r}.")
        self._log_level = log_level
        self._argument_name = argument_name

        for note in extra_notes:
            self.add_note(note)

        self.add_note("Hint: Register this level using `logging.addLevelName()`.")

    @property
    def log_level(self) -> LogLevel:
        """Log level provided by user."""
        return self._log_level

    @property
    def argument_name(self) -> str:
        """Argument name seen by user."""
        return self._argument_name


def convert_log_level(
    log_level: LogLevel,
    *,
    verify: bool = False,
    name: str = "log_level",
) -> int:
    """Convert ``str`` or ``int`` to a valid log level.

    Args:
        log_level: Either ``str`` or ``int``.
        verify: If ``True`` and `log_level` is an ``int``, verify that `log_level` is a known level.
        name: Name to use in error messages.

    Returns:
        An integer level.

    See Also:
        The :func:`logging.addLevelName` function.

    Raises:
        LogLevelError: If `log_level` is an ``int`` without a known level name and ``verify=True``.
        LogLevelError: If `log_level` is a ``str`` without a known level name.
    """
    if isinstance(log_level, int):
        if verify and log_level not in _l._levelToName:
            verify = False
            raise LogLevelError(log_level, name, extra_notes=[f"Hint: Set `{verify=}` to allow."])

        return log_level

    name_to_level = _l._nameToLevel
    int_level = name_to_level.get(log_level)
    if int_level is None:
        upper = log_level.upper()
        extra_notes = [f"Hint: Did you mean {name}={upper!r}?"] if upper in name_to_level else []
        raise LogLevelError(log_level, name, extra_notes=extra_notes)

    return int_level


_UserLoggerArg = str | _l.Logger | _l.LoggerAdapter[_t.Any]
_UserVerbosityLevel = (
    _t.Mapping[str, LogLevel]
    | _t.Mapping[_l.Logger, LogLevel]
    | _t.Mapping[str | _l.Logger, LogLevel]
    | _t.Mapping[str | _l.LoggerAdapter[_t.Any], LogLevel]
    | _t.Mapping[_UserLoggerArg, LogLevel]
)
_UserVerbosityLevels = _t.Iterable[_UserVerbosityLevel]


class LoggingSetupHelper:
    """Helper class for logging configuration.

    Verbosity:
        Verbosity levels are assumed to be strictly increasing in the amount of logs emitted. In other words, `levels`
        such as

        >>> [{"id_translation": "DEBUG"}, {"id_translation": "INFO"}]

        are **not** permitted, since going from ``verbosity=1`` to ``verbosity=2`` would reduce the amount of legts
        emitted. Calling :meth:`LoggingSetupHelper.configure_logging` with ``verbosity=0`` disables logging entirely.

    Args:
        levels: An iterable of levels, where each level is a dict ``{logger_name: log_level}``.

        format: Format string for emitted messages; see :func:`rics.logs.basic_config`.
        datefmt: Format string for date/time; see :func:`rics.logs.basic_config`.

    .. note::

       You should `not` include the ``'_level'`` suffix that you would use when calling :func:`basic_config` directly.

    Raises:
        TypeError: If any `levels` are empty.
        ValueError: If `levels` is not strictly increasing in verbosity.
    """

    def __init__(
        self,
        levels: _UserVerbosityLevels,
        *,
        format: str = FORMAT_SEC,
        datefmt: str = DATE_FORMAT,
    ) -> None:
        self._levels = self._process_user_input_levels(levels)
        self._format = format
        self._datefmt = datefmt

    @property
    def max(self) -> int:
        """Maximum verbosity level."""
        return len(self._levels)

    def get_log_levels(self, verbosity: int) -> dict[str, int]:
        """Get log levels per logger at the given `verbosity` level.

        Args:
            verbosity: Verbosity level. Pass -1 for max level.

        Returns:
            A dict ``{logger: log_level}``

        Raises:
            ValueError: If `verbose` is zero, less than -1, or too high.
        """
        max_level = -1
        if verbosity in {max_level, self.max}:
            levels = self._levels
        elif verbosity == 0:
            raise ValueError(f"{verbosity=} = 0")
        elif verbosity < max_level:
            raise ValueError(f"{verbosity=} < -2")
        else:
            try:
                levels = self._levels[:verbosity]
            except IndexError:
                msg = f"Got {verbosity=}, but only {self.max} levels are defined."
                raise ValueError(msg) from None

        rv: dict[str, int] = {}
        for level_dict in reversed(levels):
            for name, level in level_dict.items():
                rv.setdefault(name, level)
        return rv

    @classmethod
    def _on_verbosity_zero(cls) -> None:
        _l.root.disabled = True  # Prevent last resort logging to stderr.

    def configure_logging(self, verbosity: int) -> None:
        """Configure logging using :func:`basic_config`.

        Args:
            verbosity: Verbosity level. Disable ``logging.root`` if zero.
        """
        if verbosity == 0:
            self._on_verbosity_zero()
            return

        kwargs = self.get_kwargs(verbosity)
        basic_config(**kwargs)

    def get_kwargs(self, verbosity: int) -> dict[str, _t.Any]:
        """Create kwargs for :func:`basic_config`.

        Args:
            verbosity: Verbosity level.

        Returns:
            A dict.
        """
        levels = self.get_log_levels(verbosity)
        kwargs: dict[str, _t.Any] = {f"{logger_name}_level": log_level for logger_name, log_level in levels.items()}
        return {
            **kwargs,
            "level": kwargs.pop("root", None),
            "format": self._format,
            "datefmt": self._datefmt,
        }

    def get_level_descriptions(self) -> list[str]:
        """Get lines describing the configured verbosity level.

        Returns one ``str`` item per level.

        Returns:
            A list of strings.
        """
        rv = []

        known_level_names = {*_l.getLevelNamesMapping()}

        seen: dict[str, int] = {}
        for verbosity in self._by_log_level():
            line = []
            for log_level, logger_names in verbosity.items():
                new_logger_names = []
                for name in logger_names:
                    if seen.get(name) != log_level:
                        new_logger_names.append(name)
                    seen[name] = log_level

                if new_logger_names:
                    name = _l.getLevelName(log_level)
                    name = f"{name} ({log_level})" if name in known_level_names else f"<no name> ({log_level})"
                    line.append(f"{' & '.join(new_logger_names)}: {name}")

            rv.append(", ".join(line))

        return rv

    def _by_log_level(self) -> list[dict[int, list[str]]]:
        rv = []

        for verbose in range(len(self._levels)):
            level_to_loggers: dict[int, list[str]] = {}
            for logger_name, log_level in self.get_log_levels(verbose + 1).items():
                level_to_loggers.setdefault(log_level, [])
                level_to_loggers[log_level].append(logger_name)

            level_to_loggers_sorted = {
                # Levels by descending verbosity (DEBUG=10, INFO=20, etc). Loggers by name.
                log_level: sorted(level_to_loggers[log_level])
                for log_level in sorted(level_to_loggers)
            }
            rv.append(level_to_loggers_sorted)

        return rv

    @classmethod
    def _process_user_input_levels(cls, user_verbosity_levels: _UserVerbosityLevels) -> list[dict[str, int]]:
        if not user_verbosity_levels:
            msg = "No levels given."
            raise TypeError(msg)

        seen: dict[str, int] = {}

        rv = []
        for i, user_dict in enumerate(user_verbosity_levels):
            if not user_dict:
                msg = f"Level {i + 1} (index={i}): No items."
                raise TypeError(msg)

            # Convert to fixed format, e.g. {"rics.performance": logging.INFO}.
            converted: dict[str, int] = {}
            for user_logger, user_log_level in user_dict.items():
                logger = cls._get_logger_name(user_logger, i)
                log_level = cls._log_level_as_int(user_log_level, logger, i)
                converted[logger] = log_level

            # Verify increasing verbosity.
            for name, new_log_level in converted.items():
                old_log_level = seen.get(name)
                if old_log_level is None:
                    continue
                if new_log_level < old_log_level:
                    continue

                old = f"{_l.getLevelName(old_log_level)} ({old_log_level})"
                new = f"{_l.getLevelName(new_log_level)} ({new_log_level})"
                msg = (
                    f"Level {i + 1} (index={i}), logger={name!r}: Found transition '{old} -> {new}'. "
                    "Log Level must decrease with increasing verbosity."
                )
                raise ValueError(msg)

            seen.update(converted)
            rv.append(converted)

        return rv

    @classmethod
    def _get_logger_name(cls, user_logger: _UserLoggerArg, index: int) -> str:
        logger = user_logger
        while hasattr(logger, "logger"):
            logger = logger.logger  # Unwrap adapters

        if isinstance(logger, str):
            return logger
        elif isinstance(logger, _l.Logger):
            return logger.name

        msg = f"Level {index + 1} ({index=}): Cannot derive name from {user_logger=}."
        raise TypeError(msg)

    @classmethod
    def _log_level_as_int(cls, user_log_level: LogLevel, user_logger: str, index: int) -> int:
        try:
            return convert_log_level(user_log_level, name=user_logger)
        except LogLevelError as e:
            msg = f"Level {index + 1} ({index=}): {e.args[0]}"
            raise ValueError(msg) from e


def _extract_extra_levels(
    **kwargs: _t.Any,
) -> tuple[dict[str, int], dict[str, _t.Any]]:
    levels: dict[str, int] = {}
    for key in list(kwargs.keys()):
        if key and key.endswith("_level"):
            wildcard_key = key.removesuffix("_level").replace("__", ".")

            level = kwargs.pop(key)
            if level is None:  # pragma: no cover
                continue

            log_level = convert_log_level(level, name=key)
            levels[wildcard_key] = log_level

    return levels, kwargs


LoggerArg = _l.Logger | str | None | _t.Literal[False]


@_t.overload
def get_logger(logger: LoggerArg) -> _l.Logger: ...
@_t.overload
def get_logger(logger: _l.LoggerAdapter[_t.Any]) -> _l.LoggerAdapter[_t.Any]: ...
def get_logger(logger: LoggerArg | _l.LoggerAdapter[_t.Any]) -> _l.Logger | _l.LoggerAdapter[_t.Any]:
    """Get a logger.

    Returns a :attr:`~logging.Logger.disabled` logger if `logger` is ``None``. Loggers and adapters are returned as-is.

    Args:
        logger: An adapter, logger, logger name, or ``None``.

    Returns:
        A logger instance.
    """
    if logger is None:
        logger = _l.getLogger("null")
        logger.disabled = True
        return logger

    if isinstance(logger, str):
        return _l.getLogger(logger)

    if isinstance(
        logger, (_l.Logger, _l.LoggerAdapter)
    ):  # Rule is deprecated. https://github.com/astral-sh/ruff/pull/16681
        return logger

    msg = f"{logger=} is not a logging.Logger"
    raise TypeError(msg)



"""Types used by the framework."""

import typing as _t
from collections.abc import Callable as _Callable

DataType = _t.TypeVar("DataType")
"""A type of data accepted by candidate functions."""
CandFunc = _Callable[[DataType], _t.Any]
"""A function ``(DataType) -> Any`` under test. The return value is ignored."""
ResultsDict = dict[str, dict[_t.Hashable, list[float]]]
"""A result set on the form ``{candidate_label: {data_label: [runtime, ...]}}``."""

Ts = _t.TypeVarTuple("Ts")
"""Argument types for a :attr:`DataFunc` callable."""
DataFunc = _Callable[[*Ts], DataType]
"""A function ``(*Ts) -> DataType`` used to generate test data."""



"""Utilities for working with Jupyter."""

from ._kernel_helper import KernelHelper
from ._kernel_spec import KernelSpec
from ._venv_helper import VenvHelper

__all__ = [
    "KernelHelper",
    "KernelSpec",
    "VenvHelper",
]



"""Legacy module.

Schedule:
    * ``rics==7.0.0``: Drop legacy module.

Please use :mod:`rics.env.interpolation` instead.
"""

from rics.env.interpolation import UnsetVariableError, Variable, replace_in_string

__all__ = ["UnsetVariableError", "Variable", "replace_in_string"]

import os
import warnings

if os.environ.get("SPHINX_BUILD") == "true":
    # Ugly fix for duplicate indexes.
    del __all__, UnsetVariableError, Variable, replace_in_string

warnings.warn(
    f"Package `{__package__}` is deprecated. Use `rics.env.interpolation` instead.",
    UserWarning,
    stacklevel=2,
)

del warnings
del os

# TODO(7.0.0): Remove this module



from rics import strings


def format_perf_counter(start: float, end: float | None = None) -> str:  # pragma: no cover
    """Deprecated alias of :func:`rics.strings.format_perf_counter`."""
    _warn("format_perf_counter")  # TODO(7.0.0): Remove this function.
    return strings.format_perf_counter(start, end=end, full=True)


def format_seconds(t: float, *, allow_negative: bool = False) -> str:  # pragma: no cover
    """Deprecated alias of :func:`rics.strings.format_seconds`."""
    _warn("format_seconds")  # TODO(7.0.0): Remove this function.
    return strings.format_seconds(t, allow_negative=allow_negative, full=True)


def _warn(name: str) -> None:  # pragma: no cover
    import warnings

    msg = (
        f"Function `rics.performance.{name}()` is deprecated."
        f"\nUse `rics.strings.{name}()` instead, or pin `rics==4.0.1` to hide this warning."
    )
    warnings.warn(msg, category=DeprecationWarning, stacklevel=3)



import re
from dataclasses import dataclass
from os import environ
from typing import ClassVar


@dataclass(frozen=True)
class Variable:
    """Representation of an environment variable."""

    PATTERN: ClassVar[re.Pattern[str]] = re.compile(r"\${[ \t]*(\w+)[ \t]*(?::(.*))?}?}")
    """The Regex pattern used to find variables in strings."""

    name: str
    """Name of the environment variable, e.g. ``PATH`` or ``USER``."""
    default: str | None
    """Fallback value is the variable is unset."""
    full_match: str
    """The entire matching (sub)string as it were found in the input."""

    @classmethod
    def parse_first(cls, s: str) -> "Variable":
        """Return the first ``Variable`` found in `s`."""
        res = cls.parse_string(s)
        if not res:
            raise ValueError(f"No variables found in '{s}'")  # pragma: no cover
        return res[0]

    @classmethod
    def parse_string(cls, s: str) -> list["Variable"]:  # noqa: PLR0912
        """Extract a list of ``Variable``-instances from a string `s`."""
        retval = []
        prev_is_dollar_sign: bool = False
        n_open: int = 0

        state = "search"

        full_match: list[str] = []
        default: list[str] = []
        name: list[str] = []

        def make(*, make_default: bool) -> None:
            nonlocal state

            var = Variable(
                "".join(name).strip(),
                default="".join(default) if make_default else None,
                full_match="".join(full_match),
            )
            name.clear()
            default.clear()
            full_match.clear()
            state = "search"
            retval.append(var)

        for c in s:
            if state != "search":
                full_match.append(c)

            if c == "{":
                n_open += 1
            elif c == "}":
                n_open -= 1

            if state == "search":
                if prev_is_dollar_sign and c == "{":
                    full_match.append("${")
                    state = "name"
            elif state == "name":
                if c == "}":
                    make(make_default=False)
                elif c == ":":
                    state = "default"
                else:
                    name.append(c)
            elif state == "default":
                if n_open == 0:
                    make(make_default=True)
                else:
                    default.append(c)

            prev_is_dollar_sign = c == "$"

        return [var for var in retval if var.name.isidentifier()]

    @property
    def is_required(self) -> bool:
        """Variables declared with only a name are required."""
        return self.default is None

    @property
    def is_optional(self) -> bool:
        """Variables declared with a default value are optional."""
        return not self.is_required

    def get_value(self, resolve_nested_defaults: bool = False) -> str:
        """Return the value of this ``Variable``.

        Args:
            resolve_nested_defaults: If ``True``, look for nested variables within this variable's `argument`-value
                if :attr:`name` is not set.

        Returns:
            The value of :attr:`name` in the system environment, or the specified default value.

        Raises:
            UnsetVariableError: If unset with no default specified. May also be raised by inner variables.
            NotImplementedError: If an inner ``Variable``, resolved from :attr:`default`, contains text that was not
                interpolated text.
            NotImplementedError: If :attr:`default` contains multiple inner variables.

        """
        if self.name in environ:
            return environ[self.name]

        if self.default is None:
            raise UnsetVariableError(self.name)

        if resolve_nested_defaults:
            stripped = self.default.strip()
            inners = Variable.parse_string(stripped)
            if len(inners) == 1:
                inner = inners[0]
                if inner.full_match != stripped:
                    raise NotImplementedError("inner match must be stripped")

                return inner.get_value(resolve_nested_defaults=resolve_nested_defaults)

            elif len(inners) > 1:
                raise NotImplementedError(f"Multiple inner variables not supported. Got: {inners} from {self}")

        return self.default


class UnsetVariableError(ValueError):
    """Error message for unset variables."""

    def __init__(self, name: str, message: str = "Not set.") -> None:
        super().__init__(f"Required Environment Variable {name!r}: {message}")
        self.name = name



from collections.abc import Hashable, Iterable
from typing import Any, Literal, TypeGuard, cast, get_args

import pandas as pd

from rics.types import verify_literal

from .types import ResultsDict


def to_dataframe(run_results: ResultsDict, names: Iterable[str] = ()) -> pd.DataFrame:
    """Create a DataFrame from performance run output, adding derived values.

    Args:
        run_results: A dict `run_results` on the form ``{candidate_label: {data_label: [runtime, ...]}}``, returned
            by :meth:`rics.performance.MultiCaseTimer.run`.
        names: Level names for tuple keys in the data (creates new columns). See :func:`.plot_run` for details.

    Returns:
        The `run_result` input as a DataFrame.
    """
    names = tuple(names)
    frames = []
    for candidate_label, candidate_results in run_results.items():
        for data_label, data_results in candidate_results.items():
            n_runs = len(data_results)
            data = {
                "Candidate": candidate_label,
                "Run no": range(n_runs),
                "Time [s]": data_results,
                "Test data": [data_label] * n_runs,
            }

            if _has_names(data_label, names=names):
                for label_part, name in zip(data_label, names, strict=True):
                    if name in data:
                        msg = f"Bad {name=}. Key is already in use: {list(data)}."
                        raise ValueError(msg)
                    data[name] = label_part

            frame = pd.DataFrame.from_dict(data, orient="columns")
            frames.append(frame)

    df = pd.concat(frames, ignore_index=True)
    df["Time [ms]"] = df["Time [s]"] * 1000
    df["Time [μs]"] = df["Time [ms]"] * 1000
    df["Time [ns]"] = df["Time [μs]"] * 1000

    groupby = df.groupby("Test data")["Time [s]"]
    df["Times min"] = df["Time [s]"] / df["Test data"].map(groupby.min())
    df["Times mean"] = df["Time [s]"] / df["Test data"].map(groupby.mean())

    return df


def _has_names(data_label: Hashable, *, names: tuple[str, ...]) -> TypeGuard[tuple[str, ...]]:
    if len(names) == 0:
        return False

    if not isinstance(data_label, tuple):
        msg = f"Expected a tuple-key in `test_data` since {names=}."
        raise TypeError(msg)

    if len(data_label) != len(names):
        msg = f"Length of {data_label=} ({len(data_label)} does not match length of {names=} ({len(names)})."
        raise TypeError(msg)

    return True


def get_best(
    run_results: ResultsDict | pd.DataFrame,
    per_candidate: bool = False,
    names: Iterable[str] = (),
) -> pd.DataFrame:
    """Get a summarized view of the best run results for each candidate/data pair.

    Args:
        run_results: Output of :meth:`rics.performance.MultiCaseTimer.run`.
        per_candidate: If ``True``, show the best times for all candidate/data pairs. Otherwise, just show the best
            candidate per data label.
        names: Data label columns to show. Use single `'Test data'` column if not given.

    Returns:
        The best (lowest) times for each candidate/data pair.

    """
    df = run_results if isinstance(run_results, pd.DataFrame) else to_dataframe(run_results, names=names)
    return df.sort_values("Time [s]").groupby(["Candidate", "Test data"] if per_candidate else "Test data").head(1)


Unit = Literal["s", "ms", "μs", "us", "ns"]
X = Literal["candidate", "data"]


def legacy_plot_run(  # pragma: no coverage
    run_results: ResultsDict | pd.DataFrame,
    x: X | None = None,
    unit: Unit | None = None,
    **kwargs: Any,
) -> None:
    """Plot the results of a performance test.

    This is a legacy method that does not support facets.

    .. figure:: ../_images/perf_plot.png

       Comparison of ``time.sleep(t)`` and ``time.sleep(5*t)``.

    Args:
        run_results: Output of :meth:`rics.performance.MultiCaseTimer.run`.
        x: The value to plot on the X-axis, using the other to determine hue. Default=derive.
        unit: Time unit to plot on the Y-axis. Default=derive.
        **kwargs: Keyword arguments for :func:`seaborn.barplot`.

    Raises:
        ModuleNotFoundError: If Seaborn isn't installed.
        TypeError: For unknown `unit` arguments.
    """
    import warnings

    import matplotlib.pyplot as plt
    from seaborn import barplot, move_legend

    data = to_dataframe(run_results) if isinstance(run_results, dict) else run_results.copy()
    data[["Test data", "Candidate"]] = data[["Test data", "Candidate"]].astype("category")

    if x is None:
        x_arg, hue = _smaller_as_hue(data)
    else:
        verify_literal(x, X, name="x")
        x_arg, hue = ("Test data", "Candidate") if x == "data" else ("Candidate", "Test data")

    if unit is None:
        unit = _unit_from_data(data)
    else:
        verify_literal(unit, Unit, name="unit")

    y = f"Time [{unit.replace('us', 'μs')}]"

    if y not in data:
        # Unit is not one of the literals, but we still check 'data' in case someone added more units themselves.
        raise TypeError(f"Bad {unit=}; column '{y}' not present in data.")

    fig, (left, right) = plt.subplots(
        ncols=2,
        tight_layout=True,
        figsize=(8 + 4 * data.Candidate.nunique(), 7),
        sharey=True,
    )
    left.set_title("Average")
    right.set_title("Best")
    fig.suptitle("Performance", size=24)

    barplot(ax=left, data=data, x=x_arg, y=y, hue=hue, errorbar="sd", **kwargs)
    best = data.groupby(["Test data", "Candidate"], observed=True).min().reset_index()
    barplot(ax=right, data=best, x=x_arg, y=y, hue=hue, errorbar=None, **kwargs)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        move_legend(right, "upper left", bbox_to_anchor=(1, 1))
    left.get_legend().remove()


def _smaller_as_hue(data: pd.DataFrame) -> tuple[str, str]:  # pragma: no coverage
    unique = data.nunique()
    return ("Test data", "Candidate") if unique["Test data"] < unique["Candidate"] else ("Candidate", "Test data")


def _unit_from_data(df: pd.DataFrame) -> Unit:  # pragma: no coverage
    """Pick the unit with the most "human" scale; whole numbers around one hundred."""
    from numpy import log10

    prefix = "Time ["
    columns = [c for c in df.columns if c.startswith(prefix)]
    means = df.groupby(["Test data", "Candidate"], observed=True)[columns].mean()

    residuals = log10(means) - 2
    avg_residual_by_time_column = residuals.mean(axis="index")
    column = avg_residual_by_time_column.abs().idxmin()

    unit = column.removeprefix(prefix).removesuffix("]")

    assert unit in get_args(Unit)  # noqa: S101
    return cast(Unit, unit)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from seaborn import FacetGrid

from rics.performance.plot_types import Postprocessor


def make_postprocessors(kwargs: dict[str, Any]) -> list[Postprocessor]:
    """Construct postprocessors based on `kwargs`.

    Args:
        kwargs: Keyword arguments for :func:`seaborn.catplot`. May be modified.

    Returns:
        A list of postprocessors ``(FacetGrid) -> None``.
    """
    kind = kwargs["kind"]
    postprocessors: list[Postprocessor] = []

    if kind == "bar":
        # TODO(seaborn==0.13.2): log_scale makes bars vanish, so we do this instead.
        log_scale = kwargs.pop("log_scale", None)
        horizontal = kwargs["x"].startswith("Time ")
        postprocessors.append(BarPlotProcessor(bool(log_scale), horizontal))

    return postprocessors


@dataclass(frozen=True)
class BarPlotProcessor:
    log_scale: bool
    horizontal: bool

    def __call__(self, facet_grid: "FacetGrid") -> None:
        fmt = "{:.3g}" if self.log_scale else "{:.2f}"

        axes = [*facet_grid.axes.flat]
        for ax in axes:
            for c in ax.containers:
                ax.bar_label(c, fmt=fmt, padding=10, bbox={"boxstyle": "round", "fc": "0.8"})

        if self.log_scale:
            # TODO(seaborn==0.13.2): log_scale makes bars vanish, so we do this instead.
            for ax in axes:
                if self.horizontal:
                    ax.set_xscale("log")
                else:
                    ax.set_yscale("log")



import json
import logging
import subprocess
import sys
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

from rics.logs import get_logger
from rics.paths import any_path_to_path

from ..misc import tname
from . import _kernel_spec, _venv_helper


class KernelHelper:
    """Helper class for working with Jupyter kernels."""

    def __init__(
        self,
        *,
        extra_packages: Iterable[str] = (),
        variant: str | None = None,
        callback: Callable[[_kernel_spec.KernelSpec], None] | None = None,
        venv_helper: _venv_helper.VenvHelper | None = None,
        logger: logging.Logger | str = __package__,
    ) -> None:
        self.extra_packages = [p for p in map(str.strip, extra_packages) if p]
        self.venv = _venv_helper.VenvHelper() if venv_helper is None else venv_helper
        self.variant = variant
        self.callback = callback
        self.logger = get_logger(logger)

    def resolve_display_name(self, display_name: str | None = None) -> str:
        """Construct kernel display name.

        Uses `display_name` as-is if given. Keys `slug`, `variant` and `manager` are provided.
        """
        if display_name:
            return display_name.format(slug=self.venv.slug, variant=self.variant, manager=self.venv.manager)
        return f"{self.venv.slug} [{self.variant or self.venv.manager}]"

    def resolve_kernel_name(self, kernel_name: str | None = None) -> str:
        """Construct kernel name. Uses `kernel_name` as-is if given."""
        if kernel_name:
            return kernel_name

        kernel_name = f"{__package__}.{self.venv.slug}"
        if variant := self.variant:
            kernel_name = f"{kernel_name}.{variant.replace(' ', '-')}".lower()

        return kernel_name

    def install(
        self,
        user: bool | None = None,
        kernel_name: str | None = None,
        display_name: str | None = None,
        prefix: str | None = None,
        profile: str | None = None,
        env: dict[str, str] | None = None,
        frozen_modules: bool = False,
    ) -> str:
        """Install kernel spec for a virtual environment.

        Args:
            user: {user}
            kernel_name: Name of kernel. Created by :meth:`resolve_kernel_name` if ``None``.
            display_name: {display_name} Created by :meth:`resolve_display_name` if ``None``.
            prefix: {prefix}
            profile: {profile}
            env: {env}
            frozen_modules: {frozen_modules}

        Returns:
            Path where the spec was installed.

        See Also:
            Uses ``ipykernel.kernelspec.install``. See https://pypi.org/project/ipykernel/ for details.
        """
        try:
            from ipykernel.kernelspec import install
        except ModuleNotFoundError as e:
            msg = "Package `ipykernel` not installed. Run of one:\n  pip install ipykernel\n  pip install rics[cli]\nto use this feature."
            raise RuntimeError(msg) from e

        logger = self.logger
        if user is None:
            user = prefix is None
            logger.debug(f"Setting {user=} from {prefix=}.")

        if self.variant and display_name is None and (env or profile) and logger.isEnabledFor(logging.INFO):
            lines = [f"Installing variant={self.variant!r}:"]
            if env:
                lines.append(f" - Kernel variables ({len(env)}): {sorted(env)}.")
            if profile:
                lines.append(f" - Kernel profile: {profile!r}.")
            logger.info("\n".join(lines))

        path = install(
            user=user,
            kernel_name=self.resolve_kernel_name(kernel_name),
            display_name=self.resolve_display_name(display_name),
            prefix=prefix,
            profile=profile,
            env=env,
            frozen_modules=frozen_modules,
        )

        self.patch_kernel(Path(path) / "kernel.json")
        self.install_packages()

        return path

    @classmethod
    def read_kernel_spec(cls, path: str | Path) -> _kernel_spec.KernelSpec:
        path = any_path_to_path(path, postprocessor=Path.is_file)

        kernel_spec_json = path.read_text()
        kernel_spec: _kernel_spec.KernelSpec = json.loads(kernel_spec_json)

        required = {"argv", "display_name", "language"}
        if missing := required.difference(kernel_spec):
            msg = f"Not a valid {_kernel_spec.KernelSpec.__name__}: Missing required keys {sorted(missing)}."
            raise TypeError(msg)
        return kernel_spec

    @property
    def packages(self) -> list[str]:
        """List of packages to install in the kernel."""
        return ["ipykernel", *self.extra_packages]

    def patch_kernel(self, path: Path) -> None:
        """Apply kernel fixups to ensure that it actually runs in a notebook."""
        logger = self.logger

        logger.info("Patching kernel '%s' at '%s'.", self.venv.slug, path)
        kernel_spec = self.read_kernel_spec(path)
        logger.debug("Read kernel spec from path='%s':\n%s", path, kernel_spec)

        assert kernel_spec["argv"][0] == sys.executable, f"{kernel_spec['argv']=}"  # noqa: S101
        kernel_spec["argv"][0] = self.venv.executable

        metadata: dict[str, Any] = kernel_spec.setdefault("metadata", {})
        logger.debug("Updating metadata: %s.", metadata)
        self._update_metadata(metadata)

        if self.callback:
            if logger.isEnabledFor(logging.DEBUG):
                func = tname(self.callback, include_module=True)
                arg = _kernel_spec.KernelSpec.__name__
                logger.debug(f"Invoking callback: {func}({arg}).")
            self.callback(kernel_spec)

        kernel_spec_json = json.dumps(kernel_spec, indent=2, sort_keys=True)
        logger.debug("Writing kernel spec to path='%s':\n%s", path, kernel_spec_json)
        path.write_text(kernel_spec_json)

    def _update_metadata(self, metadata: dict[str, Any]) -> None:
        program_metadata: dict[str, Any] = metadata.setdefault(__package__, {})

        program_metadata["venv"] = {
            "slug": self.venv.slug,
            "version_info": self.venv.config.get("version_info"),
            "manager": self.venv.manager,
            "pyvenv.cfg": str(self.venv.config_path),
        }
        program_metadata["variant"] = self.variant
        _kernel_spec.set_metadata(program_metadata)

    def install_packages(self) -> None:
        """Installed required :attr:`additional packages <packages>`."""
        packages = self.packages
        self.logger.info("Installing packages: %s.", packages)
        flags = ["--disable-pip-version-check", "--quiet", "--require-virtualenv"]
        args = [self.venv.executable, "-m", "pip", "install", *flags, *packages]
        self.logger.debug("Installing packages: Command: %s", " ".join(args))
        self._check_call(args, timeout=180 + 20 * len(packages))

    @classmethod
    def _check_call(cls, args: list[str], timeout: float) -> None:
        subprocess.check_call(args, timeout=timeout)  # noqa: S603


DOCSTRINGS = {
    "user": "Select user [default] or system-wide install. System install may require elevated privileges.",
    "display_name": "Explicit display name.",
    "prefix": "Spec location prefix for non-default locations such as conda.",
    "profile": "An IPython profile to be loaded by the kernel.",
    "env": "Extra environment variables for the kernel.",
    "frozen_modules": "Set for potentially faster startup. Prevents debugging of built-in modules.",
}

if KernelHelper.install.__doc__:
    KernelHelper.install.__doc__ = KernelHelper.install.__doc__.format_map(DOCSTRINGS)



from typing import overload

from rics.strings import str_as_bool

from ._base import read_env


@overload
def read_bool(var: str, default: bool = False, *, strict: bool = True, split: str) -> list[bool]: ...
@overload
def read_bool(var: str, default: bool = False, *, strict: bool = True, split: None = None) -> bool: ...
def read_bool(var: str, default: bool = False, *, strict: bool = True, split: str | None = None) -> bool | list[bool]:
    """Read ``bool`` variable.

    Args:
        var: Variable to read.
        default: Default value to use when `var` is not set (or blank).
        strict: If ``False``, always fall back to `default` instead of raising.
        split: Character to split on. Returns ``list[bool]`` when set.

    Returns:
        A ``bool`` value, or a list thereof (if `split` is set).

    Notes:
        See :func:`~rics.strings.str_as_bool` for value mapping.

    Examples:
        Basic usage.

        >>> import os
        >>> os.environ["MY_BOOL"] = "true"
        >>> read_bool("MY_BOOL")
        True

        >>> os.environ["MY_BOOL"] = "0"
        >>> read_bool("MY_BOOL")
        False

        When using ``strict=False``, unmapped values are converted to the `default` instead of raising.

        >>> os.environ["MY_BOOL"] = "not-a-bool"
        >>> read_bool("MY_BOOL", default=True, strict=False)
        True

        When using `split`, elements are cleaned individually and blank items are skipped.

        >>> os.environ["MY_BOOL_LIST"] = "true, 0, no, yes, enable,, false"
        >>> read_bool("MY_BOOL_LIST", split=",")
        [True, False, False, True, True, False]

        Conversion (see :attr:`~rics.strings.str_as_bool`) must succeed for all elements, or the `default` will be
        returned.
    """
    return read_env(var, str_as_bool, default, strict=strict, split=split)



"""Performance testing utility."""

from ._format_perf_counter import format_perf_counter, format_seconds
from ._multi_case_timer import MultiCaseTimer, SkipIfParams
from ._plot import plot_run
from ._util import get_best, legacy_plot_run, to_dataframe
from ._wrapper import run_multivariate_test

__all__ = [
    "MultiCaseTimer",
    "SkipIfParams",
    "format_perf_counter",
    "format_seconds",
    "get_best",
    "legacy_plot_run",
    "plot_run",
    "run_multivariate_test",
    "to_dataframe",
]



import logging
import os
import subprocess
import sys
import tomllib
from collections.abc import Callable
from pathlib import Path
from typing import Any

from rics.logs import get_logger

Resolver = Callable[[Path, logging.Logger], tuple[str, str] | None]


class VenvHelper:
    """Helper class for working with virtual environments."""

    def __init__(
        self,
        logger: logging.Logger | str = __package__ + ".VenvHelper",
    ) -> None:
        self.logger = get_logger(logger)

        manager, exec_prefix, executable = self.resolve()
        self._manager = manager
        self.exec_prefix = exec_prefix
        self.executable = executable
        self.config_path = Path(self.exec_prefix) / "pyvenv.cfg"
        self._config = self._read_pyvenv_cfg()

    @property
    def manager(self) -> str:
        """Manager name."""
        return self._manager

    @property
    def config(self) -> dict[str, Any]:
        """Parsed `pyvenv.cfg` dict."""
        return {**self._config}

    @property
    def slug(self) -> str:
        """Virtual environment slug. Typically, the shell prompt."""
        return self._config.get("prompt") or self.config_path.parent.parent.name

    def resolve(self) -> tuple[str, str, str]:
        """Resolve venv paths.

        Returns:
            A tuple ``(name, exec_prefix, executable)``.
        """
        logger = self.logger.getChild("resolve")

        resolvers: list[Resolver] = [Resolve.poetry, Resolve.uv, Resolve.sys]
        logger.debug(f"Resolver order ({len(resolvers)}): {' | '.join(r.__name__ for r in resolvers)}.")

        cwd = Path.cwd()
        for i, resolver in enumerate(resolvers, start=1):
            name = resolver.__name__
            msg = f"Resolver {i}/{len(resolvers)}: {name}(): %s."
            logger.debug(msg, "Starting")
            result = resolver(cwd, logger)

            if result:
                logger.debug(msg, f"succeeded with {result=}")
                return name, *result

            logger.debug(msg, "Failed")

        raise RuntimeError("exec resolution failed")

    def _read_pyvenv_cfg(self) -> dict[str, str]:
        self.logger.debug("Reading pyvenv.cfg for manager='%s': '%s'.", self.manager, self.config_path)
        if not self.config_path.is_file():
            msg = f"Not a virtualenv: '{self.exec_prefix}'."
            raise RuntimeError(msg)

        lines = self.config_path.read_text().splitlines()
        rv = {key: value for key, value in map(self._convert_pyenvcfg_line, lines)}
        self.logger.debug("Found %i keys in pyvenv.cfg: %s.", len(rv), [*rv])
        return rv

    @classmethod
    def _convert_pyenvcfg_line(cls, line: str) -> tuple[str, str]:
        key, _, value = line.partition(" = ")
        return key, value


class Resolve:
    """Get ``(exec_prefix, executable)``-tuples for various package managers."""

    @classmethod
    def poetry(cls, cwd: Path, logger: logging.Logger) -> tuple[str, str] | None:
        """Implementation for https://github.com/python-poetry/poetry."""
        if not _check_files(cwd, "poetry", logger):
            return None

        return _print_paths(logger, "poetry", "run", "python")

    @classmethod
    def uv(cls, cwd: Path, logger: logging.Logger) -> tuple[str, str] | None:
        """Implementation for https://github.com/astral-sh/uv."""
        if not _check_files(cwd, "uv", logger):
            return None

        paths = _print_paths(logger, "uv", "run")

        if paths is None:
            return None

        logger.debug("Ensure pip in uv venv.")  # Not installed unless --seed/UV_VENV_SEED=pip is used.
        _get_output(logger, "uv", "pip", "install", "--no-upgrade", "--quiet", "pip")

        return paths

    @classmethod
    def sys(cls, cwd: Path, logger: logging.Logger) -> tuple[str, str]:  # noqa: ARG003
        return sys.exec_prefix, sys.executable


def _check_files(cwd: Path, name: str, logger: logging.Logger) -> bool:
    lock = cwd / f"{name}.lock"
    if lock.is_file():
        logger.debug("Found lockfile: '%s'.", lock)
        return True

    pyproject = cwd / "pyproject.toml"
    if not pyproject.is_file():
        return False
    logger.debug("Found project file: '%s'.", pyproject)

    with pyproject.open("rb") as f:
        toml = tomllib.load(f)

    rv = name in toml

    if rv:
        logger.debug("Found key=%r in project file: '%s'.", name, pyproject)

    return rv


def _print_paths(logger: logging.Logger, *args: str) -> tuple[str, str] | None:
    file = Path(__file__).parent / "_print_paths.py"
    output = _get_output(logger, *args, str(file))
    if output is None:
        return None

    exec_prefix = ""
    executable = ""
    for line in output.strip().splitlines():
        # This is a bit redundant, but helps in cases where the manager writes junk to stdout.
        if line.startswith("exec_prefix="):
            exec_prefix = line.removeprefix("exec_prefix=").rstrip()

        if line.startswith("executable="):
            executable = line.removeprefix("executable=").rstrip()

    if not (exec_prefix and executable):
        raise ValueError(f"Could not find 'exec_prefix' and 'executable' in output:\n{output}")

    return exec_prefix, executable


def _get_output(logger: logging.Logger, *args: str) -> str | None:
    logger.debug("Executing command: %s", args)

    env = {**os.environ}  # This is basically what `subprocess.run` does when env=None (the default).
    env.pop("VIRTUAL_ENV", None)  # Break out of current virtualenv. Must be unset; blank value breaks poetry.

    try:
        process = subprocess.run(args, capture_output=True, check=False, env=env)  # noqa: S603
    except OSError as e:
        _log_failed_command(logger, args, e)
        return None

    if process.returncode:
        _log_failed_command(logger, args, process)
        return None

    if logger.isEnabledFor(logging.DEBUG):
        msg = f"Command {args} succeeded:{_format_streams(process)}"
        logger.debug(msg)

    return process.stdout.decode().strip()


def _log_failed_command(
    logger: logging.Logger,
    args: tuple[str, ...],
    result: subprocess.CompletedProcess[bytes] | Exception,
) -> None:
    if not logger.isEnabledFor(logging.DEBUG):
        return

    msg = f"Command {args} failed"
    if isinstance(result, Exception):
        msg += f". {type(result).__name__}: {result}"
    else:
        msg += f" (status={result.returncode}): {_format_streams(result)}"
    logger.debug(msg)


def _format_streams(process: subprocess.CompletedProcess[bytes]) -> str:
    rows = []

    divider = "-" * 36

    if process.stderr:
        rows.append(f"\n{divider} stderr {divider}\n{process.stderr.decode().strip()}")
    if process.stdout:
        rows.append(f"\n{divider} stdout {divider}\n{process.stdout.decode().strip()}")

    return "".join(rows) if rows else " <no output>"



"""Dict utility functions."""

import typing as _t
from functools import cached_property as _cached_property
from warnings import warn as _warn

from ..action_level import ActionLevel as _ActionLevel
from ..misc import get_by_full_name as _get_by_full_name
from ..misc import tname as _tname

KT = _t.TypeVar("KT", bound=_t.Hashable)
"""Key type."""
VT = _t.TypeVar("VT")
"""Value type."""
VT_co = _t.TypeVar("VT_co", covariant=True)
"""Covariant type."""
HVT = _t.TypeVar("HVT", bound=_t.Hashable)
"""Hashable value type."""
OKT = _t.TypeVar("OKT", bound=_t.Hashable)
"""Outer key type."""


def compute_if_absent(d: dict[KT, VT], key: KT, func: _t.Callable[[KT], VT] | None = None) -> VT:
    """Compute and store `key` using `func` if `key` is not in the dict.

    Args:
        d: A dict.
        key: The key to get.
        func: A function to call for missing keys. Perform regular ``__getitem__`` call if ``None``.

    Returns:
        The value of `k` in `d`.

    """
    if not (func is None or key in d):
        d[key] = func(key)

    return d[key]


def reverse_dict(d: _t.Mapping[KT, HVT], duplicate_key_action: _ActionLevel.ParseType = "raise") -> dict[HVT, KT]:
    """Swap keys and values.

    Args:
        d: A dict to reverse.
        duplicate_key_action: Action to take if the return dict has key collisions in the reversed dict, i.e. there are
            duplicate values in `d`. Set to `ignore` to allow.

    Returns:
        A reversed copy of `d`.

    Examples:
        Reversing a dict with two elements.

        >>> reverse_dict({"A": 0, "B": 1})
        {0: 'A', 1: 'B'}

    Raises:
        ValueError: If there are duplicate values in `d` and ``duplicate_key_action='raise'``.

    """
    retval = {value: key for key, value in d.items()}

    action_level = _ActionLevel.verify(duplicate_key_action)
    if action_level is not _ActionLevel.IGNORE and len(d) != len(retval):
        msg = (
            f"Duplicate values in {d}; cannot reverse. Original dict has size {len(d)} != {len(retval)}."
            f"\nHint: Set duplicate_key_action='ignore' to allow."
        )
        if action_level is _ActionLevel.WARN:
            _warn(msg, stacklevel=2)
        else:
            raise ValueError(msg)  # pragma: no cover

    return retval


def flatten_dict(
    d: dict[KT, _t.Any],
    join_string: str = ".",
    filter_predicate: _t.Callable[[KT, _t.Any], bool] = lambda *_: True,
    string_fn: _t.Callable[[KT], str] | str | None = str,
) -> dict[str, _t.Any]:
    """Flatten a nested dictionary.

    This process is partially (or fully; depends on `d` and arguments) reversible, see :func:`unflatten_dict`.

    Args:
        d: A dict to flatten. Keys must be strings.
        join_string: Joiner for nested keys.
        filter_predicate: A callable ``(key, value) -> should_keep``. Default always returns ``True``.
        string_fn: A callable which takes a non-string key value and converts into to a string. If ``None``, a type
            error will be raised for non-string keys. Default is :py:class:`str(key) <str>`. Pass a string to resolve
            using :func:`.get_by_full_name`.

    Returns:
        A flattened version of `d`.

    Examples:
        Flattening a shallow nested dict.

        >>> flatten_dict({"foo": 0, "bar": {"foo": 1, "bar": 2}})
        {'foo': 0, 'bar.foo': 1, 'bar.bar': 2}

    """
    if string_fn is None:

        def no_string_fn(key: _t.Hashable) -> str:
            raise TypeError(f"Cannot convert {key=} with string_fn=None.")

        string_fn = no_string_fn
    elif isinstance(string_fn, str):
        string_fn = _t.cast(_t.Callable[[KT], str], _get_by_full_name(string_fn))

    retval: dict[str, _t.Any] = {}
    _flatten_inner(d, retval, [], join_string, filter_predicate, string_fn)
    return retval


def _flatten_inner(
    d: dict[KT, _t.Any],
    flattened: dict[str, _t.Any],
    parents: list[str],
    join_string: str,
    filter_predicate: _t.Callable[[KT, _t.Any], bool],
    string_fn: _t.Callable[[KT], str],
) -> None:
    for key, value in d.items():
        if not filter_predicate(key, value):
            continue

        str_key = key if isinstance(key, str) else string_fn(key)
        key_hierarchy = [*parents, str_key]
        if isinstance(value, dict):
            _flatten_inner(
                value,
                flattened,
                key_hierarchy,
                join_string,
                filter_predicate,
                string_fn,
            )
        else:
            flat_key = join_string.join(key_hierarchy)
            flattened[flat_key] = value


def unflatten_dict(
    d: dict[str | tuple[str, ...], _t.Any],
    join_string: str = ".",
) -> dict[str, dict[str, _t.Any] | _t.Any]:
    """Unflatten a flat dictionary.

    This process is reversible, see :func:`flatten_dict`.

    Args:
        d: A flat dict to unflatten. Keys must be ``str`` or ``tuple``.
        join_string: Joiner for flattened keys. Ignored if `d` has ``tuple``-keys.

    Returns:
        A nested version of `d`.

    Examples:
        Unflatten a flat dict.

        >>> unflatten_dict({"foo": 0, "bar.foo": 1, "bar.bar": 2})
        {'foo': 0, 'bar': {'foo': 1, 'bar': 2}}

        Tuple keys are also supported, including mixed key types.

        >>> unflatten_dict({"foo": 0, ("bar", "foo"): 1, "bar.bar": 2})
        {'foo': 0, 'bar': {'foo': 1, 'bar': 2}}

    """
    ret: dict[str, dict[str, _t.Any] | _t.Any] = {}
    for key, value in d.items():
        parts: _t.Sequence[str]
        if isinstance(key, str):
            parts = key.split(join_string)
        else:
            parts = key
        final = len(parts) - 1
        current = ret
        for i, p in enumerate(parts):
            if i == final:
                current[p] = value
            else:
                if p not in current:
                    current[p] = {}
                current = current[p]
    return ret


class InheritedKeysDict(_t.Mapping[OKT, dict[KT, VT]]):  # noqa: PLW1641
    """A nested dictionary that returns default-backed child dictionaries.

    The length of an ``InheritedKeysDict`` is equal to the number of specific outer keys, and is considered ``True``
    when cast to bool if there are shared and/or specific keys present.

    Args:
        default: Shared (fallback) mappings for all contexts.
        specific: Context-specific mappings, backed by the default fallback mappings.

    Examples:
        A short demonstration.

        >>> shared = {0: "fallback-for-0", 1: "fallback-for-1"}
        >>> specific = {
        ...     "ctx0": {0: "c0-v0"},
        ...     "ctx1": {0: "c1-v0", 1: "c1-v1", 2: "c1-v2"},
        ... }
        >>> ikd = InheritedKeysDict(default=shared, specific=specific)
        >>> ikd
        InheritedKeysDict(default={0: 'fallback-for-0', 1: 'fallback-for-1'}, specific={'ctx0': {0: 'c0-v0'},
        'ctx1': {0: 'c1-v0', 1: 'c1-v1', 2: 'c1-v2'}})

        The value of key `0` is inherited for `'ctx0'`. The `'ctx1'`-context
        defines all shared keys, as well as a unique key.

        >>> ikd["ctx0"]
        {0: 'c0-v0', 1: 'fallback-for-1'}
        >>> ikd["ctx1"]
        {0: 'c1-v0', 1: 'c1-v1', 2: 'c1-v2'}

        The ``InheritedKeysDict.__contains__``-method is ``True`` for all keys.
        Unknown keys simply return the default values. This will be an empty
        if no specific keys are specified.

        >>> "unseen-key" in ikd
        True
        >>> ikd["unseen-key"]
        {0: 'fallback-for-0', 1: 'fallback-for-1'}

        The length of `ikd` is equal to the number of specific contexts (two in this case).
    """

    def __init__(
        self,
        specific: dict[OKT, dict[KT, VT]] | None = None,
        default: dict[KT, VT] | None = None,
    ) -> None:
        self._specific = specific or {}
        self._default = default or {}

    def __getitem__(self, context: OKT) -> dict[KT, VT]:
        if not self:
            raise KeyError(context)

        specific = self._specific.get(context, {})
        return {**self._default, **specific}

    def __repr__(self) -> str:
        default = self._default
        specific = self._specific
        return f"{_tname(self)}({default=}, {specific=})"

    def __eq__(self, other: _t.Any) -> bool:
        if not isinstance(other, InheritedKeysDict):
            return False

        return self._default == other._default and self._specific == other._specific

    def __bool__(self) -> bool:
        return bool(self._default or self._specific)

    def __len__(self) -> int:
        return len(self._specific)

    def __iter__(self) -> _t.Iterator[OKT]:
        yield from self._specific

    def copy(self) -> "InheritedKeysDict[OKT, KT, VT]":
        """Make a copy of this ``InheritedKeysDict``."""
        return InheritedKeysDict(specific=self._specific.copy(), default=self._default.copy())

    @classmethod
    def make(cls, arg: "MakeType[OKT, KT, VT]") -> "InheritedKeysDict[OKT, KT, VT]":
        """Create instance from a mapping.

        The given argument must be on the format::

            {
                "default": {key: value},
                "specific": {
                    ctx0: {key: value},
                    ctx1: {key: value},
                    ...
                    ctxN: {key: value},
                }
            }

        No other top-level keys are accepted, but neither `default` nor `context-specific` are required.

        Args:
            arg: Input to make an instance from.

        Returns:
            A new instance.

        Raises:
            ValueError: If there are any keys other than 'default' and 'context-specific' present in `mapping`.

        """
        if isinstance(arg, InheritedKeysDict):  # pragma: no cover
            return arg

        default: dict[KT, VT] | None = arg.pop("default", None)
        specific: dict[OKT, dict[KT, VT]] | None = arg.pop("specific", None)

        if arg:  # pragma: no cover
            raise ValueError(f"Unknown keys: {list(arg)}")

        return InheritedKeysDict(default=default, specific=specific)


class MakeDict(_t.TypedDict, _t.Generic[OKT, KT, VT], total=False):
    """See :meth:`InheritedKeysDict.make`."""

    default: dict[KT, VT]
    specific: dict[OKT, dict[KT, VT]]


_MakeDict = MakeDict  # TODO(7.0.0): Remove alias.

MakeType = InheritedKeysDict[OKT, KT, VT] | MakeDict[OKT, KT, VT]
"""See :meth:`InheritedKeysDict.make`."""


def format_changed_keys(
    mapping: _t.Mapping[KT, VT_co],
    *,
    baseline: _t.Mapping[KT, VT_co],
    added: bool | _t.Literal["keys"] = True,
    updated: bool | _t.Literal["keys"] = True,
    removed: bool | _t.Literal["keys"] = "keys",
) -> str:
    """Format difference from a `baseline` mapping.

    Pass ``'keys'`` instead of ``True`` / ``False`` to show affected keys without values.

    Args:
        mapping: The current mapping.
        baseline: A baseline mapping.
        added: Determines how and if to print :attr:`~DictComparer.added` keys.
        updated: Determines how and if to print :attr:`~DictComparer.updated` keys.
        removed: Determines how and if to print :attr:`~DictComparer.removed` keys.

    Returns:
        Stylized difference between a `mapping` and the `baseline`.

    Examples:
        >>> mapping = {0: None, 1: 1, -1: None}
        >>> b = {0: 0, 1: 1, 2: 2}  # baseline

        Default behavior.

        >>> format_changed_keys(mapping, baseline=b)
        'added(1): {-1: None}, updated(1): {0: None}, removed(1): {2}'

        Showing added keys without values.

        >>> format_changed_keys(mapping, baseline=b, added="keys")
        'added(1): {-1}, updated(1): {0: None}, removed(1): {2}'

        Removed keys are printed without values by default. Passing ``True`` prints the values.

        >>> format_changed_keys(mapping, baseline=b, removed=True)
        'added(1): {-1: None}, updated(1): {0: None}, removed(1): {2: 2}'

        To hide a change type entirely, pass ``False`` instead.

        >>> format_changed_keys(mapping, baseline=b, updated=False)
        'added(1): {-1: None}, removed(1): {2}'

    Notes:
        This is a convenience function that warps :func:`DictComparer.to_string`.
    """
    return DictComparer(mapping, baseline=baseline).to_string(added=added, updated=updated, removed=removed)


_ChangesByType = dict[str, set[KT] | dict[KT, VT_co]]


class DictComparer(_t.Generic[KT, VT_co]):
    """Utility class for comparing mapping types.

    Args:
        mapping: The current mapping.
        baseline: A baseline mapping.
        preference: Determines values in :attr:`updated`. Set to `'old'` to prefer `baseline` values.

    Notes:
        * Properties are cached.
        * ``DictComparer.__str__`` is :meth:`DictComparer.to_string`.
    """

    Preference = _t.Literal["old", "new"]
    """Value preference for :attr:`updated` keys."""

    def __init__(
        self,
        mapping: _t.Mapping[KT, VT_co],
        *,
        baseline: _t.Mapping[KT, VT_co],
        preference: Preference = "new",
    ) -> None:
        from rics.types import verify_literal

        self._baseline = baseline
        self._mapping = {} if mapping is None else mapping

        verify_literal(preference, DictComparer.Preference, name="preference")
        self._prefer_new = preference == "new"

    def to_string(
        self,
        *,
        added: bool | _t.Literal["keys"] = True,
        updated: bool | _t.Literal["keys"] = True,
        removed: bool | _t.Literal["keys"] = "keys",
        formatter: _t.Callable[[_ChangesByType[KT, VT_co]], str] | None = None,
    ) -> str:
        """Format :attr:`changed` keys.

        Pass ``'keys'`` instead of ``True`` / ``False`` to show affected keys without values.

        Args:
            added: Determines how and if to print :attr:`added` keys.
            updated: Determines how and if to print :attr:`updated` keys.
            removed: Determines how and if to print :attr:`removed` keys.
            formatter: A callable ``({change_type: {key: value} | {key, ...}) -> str``, where
                ``change_type = 'added'|'updated'|'removed'``. Use :meth:`default_formatter` if ``None``.

        Returns:
            Pretty-printed changes.
        """
        if not (added or updated or removed):
            msg = "At least one of {'added', 'updated', 'removed'} must be set"
            raise TypeError(msg)

        changes_by_type: _ChangesByType[KT, VT_co] = {}
        if added and (added_ := self.added):
            changes_by_type["added"] = added_ if added is True else set(added_)
        if updated and (updated_ := self.updated):
            changes_by_type["updated"] = updated_ if updated is True else set(updated_)
        if removed and (removed_ := self.removed):
            changes_by_type["removed"] = removed_ if removed is True else set(removed_)

        if formatter is None:
            formatter = self.default_formatter
        return formatter(changes_by_type)

    @property
    def baseline(self) -> _t.Mapping[KT, VT_co]:
        """Get the original `baseline` mapping."""
        return self._baseline

    @property
    def mapping(self) -> _t.Mapping[KT, VT_co]:
        """Get the original `mapping`."""
        return self._mapping

    @property
    def preference(self) -> Preference:
        """Value preference for :attr:`updated` key."""
        return "new" if self._prefer_new else "old"

    __str__ = to_string

    def __repr__(self) -> str:
        mapping = self.mapping
        baseline = self.baseline
        preference = self.preference
        return f"{type(self).__name__}({mapping=}, {baseline=}, {preference=})"

    @property
    def total(self) -> int:
        """Total number of :attr:`changed` keys."""
        return len(self.changed)

    @_cached_property
    def changed(self) -> dict[KT, VT_co]:
        """Get a dict containing all :attr:`added`, :attr:`changed` or :attr:`removed` keys."""
        return {**self.added, **self.updated, **self.removed}

    @_cached_property
    def added(self) -> dict[KT, VT_co]:
        """Get a dict containing new keys."""
        return {k: v for k, v in self._mapping.items() if k not in self._baseline}

    @_cached_property
    def removed(self) -> dict[KT, VT_co]:
        """Get a dict containing removed keys."""
        return {k: v for k, v in self._baseline.items() if k not in self._mapping}

    @_cached_property
    def updated(self) -> dict[KT, VT_co]:
        """Get a dict containing updated values (see :attr:`preference`) for :attr:`baseline` keys."""
        baseline = self._baseline
        mapping = self._mapping

        updated = {}
        for key in set(baseline).intersection(mapping):
            old = baseline[key]
            new = mapping[key]

            if old != new:
                # if self._recursive and isinstance(new, _t.Mapping) and isinstance(old, _t.Mapping):
                #    cls = type(self)
                #    updated_child = cls(new, baseline=old, preference=self.preference, recursive=True).updated
                #    updated[key] = updated_child
                # else:
                updated[key] = new if self._prefer_new else old

        return updated

    @classmethod
    def default_formatter(cls, changes_by_type: _ChangesByType[_t.Any, _t.Any]) -> str:
        """Default formatting implementation."""
        return ", ".join(
            f"{change_type}({len(changes):_d}): {changes!r}"  # count and type
            for change_type, changes in changes_by_type.items()
        )



import functools
import logging
import warnings
from collections.abc import Callable, Collection, Hashable, Iterator, Mapping
from dataclasses import dataclass
from inspect import Parameter, signature
from time import perf_counter
from timeit import Timer
from typing import Any, ClassVar, Generic, TypeAlias

from rics.misc import format_kwargs, tname
from rics.strings import format_perf_counter
from rics.strings import format_seconds as fmt_time

from .types import CandFunc, DataFunc, DataType, ResultsDict, Ts

UNRELIABLE_RESULTS_LIMIT = 1e-6  # Prevent spurious "4x" warnings.

CandidateMethodArg: TypeAlias = Mapping[str, CandFunc[DataType]] | Collection[CandFunc[DataType]] | CandFunc[DataType]
TestDataArg: TypeAlias = Mapping[Any, DataType] | Collection[DataType]


class MultiCaseTimer(Generic[DataType, *Ts]):
    """Performance testing implementation for multiple candidates and data sets.

    Test data:
        * Typically a dict ``{label: data}`` to evaluate candidates.

          * Other collections are converted to ``dict`` using :meth:`process_test_data`. String label will then be based
            on sample data.
          * Labels may also be ``tuple``. This may then be used to plot different categories of data in different
            facets; see the :func:`.plot_run` function with the `names` argument.
          * For non-dict inputs, string labels will be generated automatically.

        * If `test_data` is :py:func:`callable`, test data will be generated from the `case_args`.

          * The `case_args` will be passed as positional arguments.
          * The `case_args` will be used as the output labels when using :meth:`run` (similar to the ``setup`` option
            provided by the built-in :py:mod:`timeit` module).

        Data access time is *not* measured by the ``run`` method.

    Args:
        candidate_method: A dict ``{label: function}``. Alternatively, you may pass a collection of functions or a
            single function.
        test_data: A ``{label: data}`` to evaluate candidates on. You may also pass a list of data, which will be
            converted to a dict as above. Data may also be generated by passing a callable.
        case_args: These are positional arguments for the `test_data` callable.
        kwargs: Shared keyword arguments for the `test_data` callable.

    Raises:
        TypeError: If `args` or `kwargs` are set when `test_data` is not a callable.
        ValueError: If `args` is empty and `test_data` is a callable.
    """

    LOGGER: ClassVar[logging.Logger | logging.LoggerAdapter[Any]] = logging.getLogger(__package__)
    """Class logger instance."""

    def __init__(
        self,
        candidate_method: CandidateMethodArg[DataType],
        test_data: TestDataArg[DataType] | DataFunc[*Ts, DataType],  # DataFunc[DataFuncP, DataType]
        *,
        case_args: Collection[tuple[*Ts]] | None = None,
        kwargs: Any | None = None,
    ) -> None:
        self._candidates = self.process_candidates(candidate_method)

        self._data: dict[Hashable, DataType] | GeneratedData[DataType, *Ts]
        if callable(test_data):
            if not case_args:
                raise ValueError("No case data given.")
            cases = [*case_args]
            if len(cases) != len({*cases}):
                raise ValueError("Cases are not unique.")

            self._data = GeneratedData(test_data, cases=cases, kwargs=kwargs, logger=self.LOGGER)
        else:
            if case_args or kwargs:
                msg = "Cannot pass `case_args` or `kwargs` when `test_data` is not a callable."
                raise TypeError(msg)

            self._data = self.process_test_data(test_data)

    @classmethod
    def process_candidates(cls, candidates: CandidateMethodArg[DataType]) -> dict[str, CandFunc[DataType]]:
        """Convert input candidates to the internal format."""
        rv = cls._process_candidates(candidates)
        if rv:
            return rv
        raise ValueError("No candidates given.")  # pragma: no cover

    @classmethod
    def process_test_data(cls, test_data: TestDataArg[DataType]) -> dict[Hashable, DataType]:
        """Convert input test data to the internal format."""
        rv = {**test_data} if isinstance(test_data, Mapping) else cls._dict_from_collection(test_data)
        if rv:
            return rv
        raise ValueError("No case data given.")  # pragma: no cover

    def derive_names(self) -> list[str]:
        """Derive names argument.

        Raises:
            TypeError: If `test_data` is not callable.
        """
        if not isinstance(self._data, GeneratedData):
            raise TypeError("Cannot derive names without callable `test_data`.")

        return self._data.derive_names()

    @property
    def is_data_generated(self) -> bool:
        """Returns ``True`` if the `test_data` is callable."""
        return isinstance(self._data, GeneratedData)

    def run(
        self,
        *,
        time_per_candidate: float = 6.0,
        repeat: int = 5,
        number: int | None = None,
        skip_if: Callable[["SkipIfParams[DataType, *Ts]"], bool] | None = None,
        progress: bool = False,
    ) -> ResultsDict:
        """Run for all cases.

        Note that the test case variant data isn't included in the expected runtime computation, so increasing the
        amount of test data variants (at initialization) will reduce the amount of times each candidate is evaluated.

        Args:
            time_per_candidate: Minimum runtime per repetition and candidate label. Ignored if `number` is set.
            repeat: Number of times to repeat for all candidates per data label.
            number: Number of times to execute each candidate, per repetition.
            skip_if: A callable ``(skip_if) -> bool``; see the :class:`params <SkipIfParams>` type.
            progress: If ``True``, display a progress bar. Requires ``tqdm``.

        Examples:
            If `repeat=5` and `time_per_candidate=3` for an instance with and 2 candidates, the total
            runtime will be approximately ``5 * 3 * 2 = 30`` seconds.

        Returns:
            A dict `run_results` on the form ``{candidate_label: {data_label: [runtime, ...]}}``.

        Raises:
            ValueError: If the total expected runtime exceeds `max_expected_runtime`.

        Notes:
            * Precomputed runtime is inaccurate for functions where a single call are longer than `time_per_candidate`.

        See Also:
            The :py:class:`timeit.Timer` class which this implementation depends on.

        """
        logger = self.LOGGER
        n_cand = len(self._candidates)
        n_data = len(self._data)
        total = n_cand * n_data

        logger.debug("Begin evaluating %i combinations: %i candidates and %i test cases.", total, n_cand, n_data)

        per_candidate_number = self._compute_number_of_iterations(number, repeat, time_per_candidate, progress)

        if progress:
            from tqdm.auto import tqdm

            pbar = tqdm(total=total)
        else:
            pbar = None

        i = 0
        run_results: ResultsDict = {}
        for candidate_label, func in self._candidates.items():
            candidate_number, candidate_est_time = per_candidate_number[candidate_label]
            run_results[candidate_label] = candidate_results = {}

            logger.info(f"Evaluate candidate {candidate_label!r} {repeat}x{candidate_number} times per datum..")
            for data_label, test_data in self._data.items():
                i += 1
                if pbar:
                    pbar.desc = f"{candidate_label}({data_label})"
                    pbar.refresh()

                if skip_if:
                    skip_if_params: SkipIfParams[DataType, *Ts] = SkipIfParams(
                        candidate=func,
                        candidate_label=candidate_label,
                        data=test_data,
                        data_label=data_label,
                        est_time=None if candidate_est_time is None else candidate_est_time * repeat,
                        results_so_far=run_results,
                    )

                    if skip_if(skip_if_params):
                        if pbar:
                            pbar.update()
                        logger.debug(f"Skip combination {i}/{total}: {candidate_label!r} @ {data_label!r}.")
                        continue

                logger.debug(f"Start evaluating combination {i}/{total}: {candidate_label!r} @ {data_label!r}.")

                raw_timings = self._get_raw_timings(func, test_data, repeat, candidate_number)

                timings = [dt / candidate_number for dt in raw_timings]

                # Same heuristic as the IPython cell magic.
                best = min(timings)
                worst = max(timings)
                if best > 0 and worst >= best * 4 and worst > UNRELIABLE_RESULTS_LIMIT:
                    t = (candidate_label, data_label)
                    warnings.warn(
                        f"Results may be unreliable for {t}. The worst time {fmt_time(worst)} "
                        f"was ~{worst / best:.1f} times slower than the best time ({fmt_time(best)}).",
                        UserWarning,
                        stacklevel=1,
                    )

                candidate_results[data_label] = timings
                if pbar:
                    pbar.update()

        return run_results

    @staticmethod
    def _get_raw_timings(func: CandFunc[DataType], test_data: DataType, repeat: int, number: int) -> list[float]:
        """Exists so that it can be overridden for testing."""
        return Timer(lambda: func(test_data)).repeat(repeat, number)

    def _compute_number_of_iterations(
        self,
        number: int | None,
        repeat: int,
        time_allocation: float,
        progress: bool,
    ) -> dict[str, tuple[int, float]] | dict[str, tuple[int, None]]:
        logger = self.LOGGER

        if isinstance(number, int):
            return {c: (number, None) for c in self._candidates}

        logger.debug("Computing number of iterations with repeat=%i and time_allocation=%f.", repeat, time_allocation)
        start = perf_counter()

        if progress:
            from tqdm.auto import tqdm

            pbar = tqdm(total=len(self._candidates))
        else:
            pbar = None

        candidate_numbers = {}
        for label in self._candidates:
            if pbar:
                pbar.desc = f"autorange('{label}')"
                pbar.refresh()

            number, time = self._autonumber(time_allocation, label)
            candidate_numbers[label] = number, time
            logger.debug("Candidate number for candidate=%r: %i (time=%f).", label, number, time)
            if pbar:
                pbar.update()

        if pbar:
            pbar.clear()

        logger.info(
            f"Computed shared number for {len(candidate_numbers)} candidates in {format_perf_counter(start)}: "
            f"{candidate_numbers}."
        )

        return candidate_numbers

    def _autonumber(self, time_allocation: float, candidate_label: str) -> tuple[int, float]:
        """Based on Timer.autorange()."""
        i = 1
        while True:
            for j in 1, 2, 3, 5:
                number = i * j

                total_time_taken = 0.0
                for data in self._data.values():
                    func = self._candidates[candidate_label]
                    partial = functools.partial(func, data)
                    total_time_taken += Timer(partial).timeit(number)

                if total_time_taken >= time_allocation:
                    if total_time_taken > 1:
                        total_time_taken = round(total_time_taken, 2)
                    return number, total_time_taken
            i *= 10

    @staticmethod
    def _process_candidates(candidates: CandidateMethodArg[DataType]) -> dict[str, CandFunc[DataType]]:
        if isinstance(candidates, Mapping):
            return {**candidates}
        if callable(candidates):
            return {tname(candidates, prefix_classname=True): candidates}

        def make_label(a: Any) -> str:
            name = tname(a, prefix_classname=True)
            return name.removeprefix("candidate_")

        labeled_candidates = {make_label(c): c for c in candidates}
        if len(labeled_candidates) != len(candidates):
            raise ValueError(
                f"Derived names for input {candidates=} are not unique. Use a dict to assign candidate names."
            )
        return labeled_candidates

    @staticmethod
    def _dict_from_collection(test_data: Collection[DataType]) -> dict[Hashable, DataType]:
        result: dict[Hashable, DataType] = {}
        for data in test_data:
            s = str(data)

            if isinstance(data, (bool, float, int, str, tuple)):
                key = s
            else:
                key = f"{s[:29]}..." if len(s) > 32 else s  # noqa: PLR2004
                key = f"Sample data: '{key}'"
            result[key] = data
        return result


class GeneratedData(Generic[DataType, *Ts]):
    def __init__(
        self,
        func: DataFunc[*Ts, DataType],
        cases: list[tuple[*Ts]],
        *,
        kwargs: dict[str, Any] | None,
        logger: logging.Logger | logging.LoggerAdapter[Any],
    ) -> None:
        self._func = func
        self._cases = cases
        self._kwargs = kwargs or {}

        if hasattr(logger, "getChild"):
            logger = logger.getChild("data")

        self._logger = logger

    def derive_names(self) -> list[str]:
        case_args = self._cases
        n = len(case_args[0])
        parameters = [
            name
            for name, parameter in signature(self._func).parameters.items()
            if parameter.kind in {Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD}
        ]
        if len(parameters) < n:
            msg = (
                f"Could not derive names for test_data="
                f"{tname(self._func, prefix_classname=True)}({', '.join(parameters)}, ...)."
                f" Expected at least {n} positional parameters since {case_args[0]=}."
                f" Got {len(parameters)}: {parameters}."
            )
            raise RuntimeError(msg)
        return parameters[:n]

    def __len__(self) -> int:
        return len(self._cases)

    def keys(self) -> list[tuple[*Ts]]:
        return self._cases

    def values(self) -> Collection[DataType]:
        return [data for _, data in self.items()]

    def items(self) -> Iterator[tuple[tuple[*Ts], DataType]]:
        logger = self._logger

        kwargs = self._kwargs
        func = self._func

        logger_enabled = logger.isEnabledFor(logging.DEBUG)
        if logger_enabled:
            name = tname(func, prefix_classname=True)
            total = len(self._cases)

        for n, case in enumerate(self._cases, start=1):
            data = func(*case, **kwargs)

            if logger_enabled:
                args = ", ".join(map(repr, case))
                if kwargs:
                    args = args + ", " + format_kwargs(kwargs)
                logger.debug(f"Yield case {n}/{total}: {name}({args})={tname(data, prefix_classname=True)}.")

            yield case, data


@dataclass(frozen=True)
class SkipIfParams(Generic[DataType, *Ts]):
    """Data type for a `skip_if` predicate."""

    candidate: CandFunc[DataType]
    """Current candidate function."""
    candidate_label: str
    """Candidate label."""
    data: DataType
    """Current data."""
    data_label: Hashable | tuple[*Ts]
    """Data label."""
    est_time: float | None
    """Estimated time to finish all repetitions. Only when `number` is derived."""
    results_so_far: ResultsDict
    """A snapshot timing values."""



"""Miscellaneous utility methods for Python applications."""

import typing as _t
from importlib import import_module as _import_module
from pathlib import Path as _Path
from pprint import saferepr as _safe_repr
from types import MethodType as _MethodType
from types import ModuleType as _ModuleType

from . import paths as _paths
from .types import AnyPath as _AnyPath


def interpolate_environment_variables(
    s: str,
    *,
    allow_nested: bool = True,
    allow_blank: bool = False,
) -> str:
    """Alias of :func:`rics.env.interpolation.replace_in_string`."""
    import warnings

    from rics.env.interpolation import replace_in_string

    warnings.warn(
        "Function `interpolate_environment_variables()` is deprecated."
        "Use `rics.env.interpolation.replace_in_string` instead.",
        UserWarning,
        stacklevel=2,
    )
    # TODO(7.0.0): Remove this function

    return replace_in_string(s, allow_nested=allow_nested, allow_blank=allow_blank)


GBFNReturnType = _t.TypeVar("GBFNReturnType")
"""Output type for :func:`get_by_full_name` when using one of `instance_of` and `subclass_of`."""


@_t.overload
def get_by_full_name(
    name: str,
    default_module: str | _ModuleType | object = ...,
    *,
    instance_of: _t.Literal[None] = None,
    subclass_of: _t.Literal[None] = None,
) -> _t.Any:
    pass


@_t.overload
def get_by_full_name(
    name: str,
    default_module: str | _ModuleType | object = ...,
    *,
    instance_of: type[GBFNReturnType],
    subclass_of: _t.Literal[None] = None,
) -> GBFNReturnType:
    pass


@_t.overload
def get_by_full_name(
    name: str,
    default_module: str | _ModuleType | object = ...,
    *,
    instance_of: _t.Literal[None] = None,
    subclass_of: type[GBFNReturnType],
) -> type[GBFNReturnType]:
    pass


def get_by_full_name(
    name: str,
    default_module: str | _ModuleType | object | None = None,
    *,
    instance_of: type[GBFNReturnType] | None = None,
    subclass_of: type[GBFNReturnType] | None = None,
) -> _t.Any:
    """Combine :py:func:`importlib.import_module` and :py:func:`getattr` to retrieve items by name.

    You may retrieve top-level module members such as classes by adding the class name after the module path (e.g.
    ``logging.Logger``). The member may also be specifying using the ``':'``-syntax, e.g. ``logging:Logger``. The
    ``':'``-syntax also supports retrieving object members e.g. ``logging:Logger.info``. This is not possible using the
    dot-based syntax since ``logging.Logger`` cannot be imported by the ``importlib.import_module`` function.

    Args:
        name: A name or fully qualified name.
        default_module: A namespace to search if `name` does not contain any ``'.'`` or ``':'`` characters.
        instance_of: If given, perform :py:func:`isinstance` check on `name`.
        subclass_of: If given, perform :py:func:`issubclass` check on `name`.

    Returns:
        The object specified by `name`.

    Raises:
        ValueError: If `name` does not contain any dots and ``default_module=None``.
        ValueError: If both `instance_of` and `subclass_of` are given.
        TypeError: If an ``isinstance`` or ``issubclass`` check fails.

    Examples:
        Retrieving a ``numpy`` function by name.

        >>> get_by_full_name("numpy.isnan")
        <ufunc 'isnan'>

        Validating the return type. In the example below, we ensure that ``logging.INFO`` really is an ``int``, and that
        the :py:class:`logging.Logger` class inherits from ``logging.Filterer``.

        >>> import logging
        >>> get_by_full_name("logging.INFO", instance_of=int)
        20
        >>> get_by_full_name("logging.Logger", subclass_of=logging.Filterer)
        <class 'logging.Logger'>

        Default namespaces may be specified using the `default_module` keyword argument.

        >>> get_by_full_name("Logger", default_module="logging")
        <class 'logging.Logger'>

        The default namespace doesn't have to be a module.

        >>> get_by_full_name("info", default_module="logging.Logger").__qualname__
        'Logger.info'

        To retrieve an attribute of anything other than a module (e.g. a class member), you must use the ``:``-syntax to
        separate the module path from the attribute path.

        >>> get_by_full_name("logging:Logger.info").__qualname__
        'Logger.info'

        When using this syntax, the path to the left of the separator must be a valid module path.
    """
    name = name.strip()
    if name == "":
        msg = "Name must not be empty."
        raise ValueError(msg)

    if not (instance_of is None or subclass_of is None):
        msg = f"At least one of ({instance_of=}, {subclass_of=}) must be None."
        raise ValueError(msg)

    obj = _get_by_full_name(name, default_module=default_module)

    if instance_of is not None:  # noqa: SIM102 # See https://github.com/nedbat/coveragepy/issues/509
        if not isinstance(obj, instance_of):
            msg = f"Expected an instance of {instance_of.__name__}, but got {obj=}."
            raise TypeError(msg)

    if subclass_of is not None:
        reason = ""
        try:
            if not issubclass(obj, subclass_of):
                reason = "does not inherit from {subclass_of}"
        except TypeError as e:
            if "must be a class" in str(e):
                reason = "is not a class"
            else:
                raise
        if reason:
            pretty = tname(subclass_of, prefix_classname=True)
            reason = reason.format(subclass_of=pretty)
            msg = f"Expected a subclass of {pretty}, but {obj=} {reason}."
            raise TypeError(msg)

    return obj


def _get_by_full_name(name: str, *, default_module: str | object | None = None) -> _t.Any:
    path: list[str]

    force_default_module = name[0] == ":"

    if force_default_module and not default_module:
        msg = f"Cannot use {name=} (with leading ':') without a default module."
        raise ValueError(msg)

    root: object
    if ":" in name and not force_default_module:
        module_name, _, member = name.rpartition(":")
        root = _import_module(module_name)
        path = member.split(".")
    elif "." in name and not force_default_module:
        module_name, _, member = name.rpartition(".")
        root = _import_module(module_name)
        path = [member]
    else:
        if not default_module:
            msg = "Name must be fully qualified when no default module is given."
            raise ValueError(msg)

        if isinstance(default_module, str):
            if ":" in default_module or "." in default_module:
                root = _get_by_full_name(default_module)
            else:
                root = _import_module(default_module)
        else:
            root = default_module

        path = name.removeprefix(":").split(".")

    obj = root
    for attr in path:
        obj = getattr(obj, attr)
    return obj


def get_public_module(obj: _t.Any, resolve_reexport: bool = False, include_name: bool = False) -> str:
    """Get the public module of `obj`.

    Args:
        obj: An object to resolve a public module for.
        resolve_reexport: If ``True``, traverse the module hierarchy and look for the earliest where `obj` is
            reexported. This may be expensive.
        include_name: If ``True``, include the name of `obj` reexported from a parent module. The first instance found
            will be used if `obj` is reexported multiple times.

    Returns:
        Public module of `obj`.

    Examples:
        Public module of ``pandas.DataFrame``.

        >>> from pandas import DataFrame as obj
        >>> get_public_module(obj)
        'pandas.core.frame'
        >>> get_public_module(obj, resolve_reexport=True)
        'pandas'
        >>> get_public_module(obj, resolve_reexport=True, include_name=True)
        'pandas.DataFrame'

    Raises:
        ValueError: If `include_name` is given without `resolve_reexport`.

    See Also:
        The analogous :func:`get_by_full_name`-function.

    """
    import inspect

    if include_name and not resolve_reexport:
        raise ValueError(f"Cannot combine {include_name=} with {resolve_reexport=}.")

    module_name = obj.__module__
    assert isinstance(module_name, str), f"{module_name=}"  # noqa: S101

    parts = []
    for part in module_name.split("."):
        if part[0] == "_":
            break
        parts.append(part)

    if resolve_reexport:
        obj_id = id(obj)

        for i in range(1, len(parts) + 1):
            module = _import_module(".".join(parts[:i]))
            for name, _ in inspect.getmembers(module, predicate=lambda member: id(member) == obj_id):
                parts = parts[:i]
                if include_name:
                    parts.append(name)
                return ".".join(parts)

    if not parts:
        return module_name

    return ".".join(parts)


def tname(
    arg: type[_t.Any] | _t.Any | None,
    prefix_classname: bool = False,
    attrs: str | _t.Iterable[str] | None = "func",
    include_module: bool = False,
) -> str:
    """Get name of method or class.

    Args:
        arg: Something get a name for.
        prefix_classname: If ``True``, prepend the class name if `arg` belongs to a class.
        attrs: Attribute names to search for wrapped functions. The default, `'func'`, is the name used by the built-in
            :py:func:`functools.partial` wrapper. May cause infinite recursion.
        include_module: If ``True``, prepend the public module (see :func:`get_public_module`).

    Returns:
        A name for `arg`.

    Raises:
        ValueError: If no name could be derived for `arg`.

    """
    if arg is None:
        return "None"

    if include_module:
        module = _get_module(arg)
        name = tname(arg, prefix_classname=prefix_classname, attrs=attrs, include_module=False)
        return f"{module}.{name}"

    if attrs:
        from rics.collections.misc import as_list

        attrs = as_list(attrs)
        for attr in attrs:
            wrapped = getattr(arg, attr, arg)
            if wrapped is arg:
                break
            return tname(wrapped, prefix_classname=prefix_classname, attrs=attrs)

    return _get_name(arg, prefix_classname)


def _get_name(arg: type[_t.Any] | _t.Any, prefix_classname: bool) -> str:
    if prefix_classname and hasattr(arg, "__self__"):
        arg_self = arg.__self__
        if isinstance(arg_self, type):
            cls_name = arg_self.__name__
        elif isinstance(arg, _MethodType):
            cls_name = type(arg_self).__qualname__
        else:
            cls_name = type(arg).__name__
        return f"{cls_name}.{arg.__name__}"

    if hasattr(arg, "__qualname__"):
        return arg.__qualname__ if prefix_classname else arg.__name__
    if hasattr(arg, "fget"):
        # Instance-level properties accessed using the class.
        return tname(arg.fget, prefix_classname=prefix_classname)
    if hasattr(arg, "__name__"):
        return arg.__name__
    if hasattr(arg, "__class__"):
        return arg.__class__.__qualname__ if prefix_classname else arg.__class__.__name__
    else:
        raise ValueError(f"Could not derive a name for {arg=}.")  # pragma: no cover


def _get_module(arg: type[_t.Any] | _t.Any) -> str:
    if hasattr(arg, "fget"):
        # Instance-level properties accessed using the class.
        obj = arg.fget
    elif hasattr(arg, "__self__"):
        obj = arg.__self__
    elif hasattr(arg, "__name__"):
        obj = arg
    else:
        obj = type(arg)

    return get_public_module(obj, include_name=False, resolve_reexport=True)


def format_kwargs(kwargs: _t.Mapping[str, _t.Any], *, max_value_length: int = 80) -> str:
    """Format keyword arguments.

    .. warning::

       This function will be removed in a future version. Use :func:`.strings.format_kwargs` instead.

    Args:
        kwargs: Arguments to format.
        max_value_length: Replace value with class name above this limit. 0=no limit.

    Returns:
        A string on the form `'key0=repr(value0), key1=repr(value1)'`.

    Raises:
        ValueError: For keys in `kwargs` that are not valid Python argument names.

    Examples:
        Basic usage.

        >>> format_kwargs({"an_int": 1, "a_string": "Hello!"})
        "an_int=1, a_string='Hello!'"
    """
    # TODO(7.0.0): deprecate in favor of string.format_kwargs() + update docstring ::warning

    invalid = [k for k in kwargs if not k.isidentifier()]
    if invalid:
        raise ValueError(f"Got {len(invalid)} invalid identifiers: {invalid}.")

    def repr_value(value: _t.Any) -> str:
        value_repr = _safe_repr(value)
        if max_value_length <= 0 or len(value_repr) <= max_value_length:
            return value_repr
        return tname(value)

    return ", ".join(f"{k}={repr_value(v)}" for k, v in kwargs.items())


def get_local_or_remote(
    file: _AnyPath,
    *,
    remote_root: _AnyPath,
    local_root: _AnyPath = ".",
    force: bool = False,
    postprocessor: _t.Callable[[str], _t.Any] | None = None,
    show_progress: bool | None = None,
) -> _Path:
    r"""Retrieve the path of a local file, downloading it if needed.

    If `file` is not available at the local root path, it will be downloaded using `requests`_. A postprocessor may
    be given in which case the name of the final file will be ``local_root/<name-of-postprocessor>/file``. Removing
    a raw local file (i.e. ``local_root/file``) will invalidate postprocessed files as well.

    Args:
        file: A file to retrieve or download.
        remote_root: Remote URL where the data may be retrieved using ``requests.get``.
        local_root: Local directory where the file may be cached.
        force: If ``True``, always download and apply processing (if applicable). Existing files will be overwritten.
        postprocessor: A function which takes a single argument `input_path` and returns a pickleable type.
        show_progress: If ``True``, show a progress bar. Requires the `tqdm`_ package. If ``None``, use only if TQDM
            is installed.

    Returns:
        An absolute path to the data.

    Raises:
        ValueError: If local root path does not exist or is not a directory.
        ValueError: If the local file does not exist and ``remote=None``.
        ModuleNotFoundError: If the ``tqdm`` package is not installed but ``show_progress=True``.

    Examples:
        Fetch the Title Basics table (a CSV file) of the `IMDb dataset`_.

        >>> from rics.misc import get_local_or_remote
        >>> import pandas as pd
        >>>
        >>> file = "name.basics.tsv.gz"
        >>> local_root = "my-data"  # default = "."
        >>> remote_root = "https://datasets.imdbws.com"
        >>> path = get_local_or_remote(
        ...     file, remote_root, local_root, show_progress=True
        ... )  # doctest: +SKIP
        >>> pd.read_csv(path, sep="\t").shape  # doctest: +SKIP
        https://datasets.imdbws.com/name.basics.tsv.gz: 100%|██████████| 214M/214M [00:05<00:00, 39.3MiB/s]
        (11453719, 6)

        We had download `name.basics.tsv.gz` the first time, but ``get_local_or_remote`` returns immediately the second
        time it is called. Fetching can be forced using ``force_remote=True``.

        >>> path = get_local_or_remote(
        ...     file, remote_root, local_root, show_progress=True
        ... )  # doctest: +SKIP
        >>> pd.read_csv(path, sep="\t").shape  # doctest: +SKIP
        (11453719, 6)

    .. _IMDb dataset:
        https://www.imdb.com/interfaces/
    .. _requests:
        https://requests.readthedocs.io/en/latest/api/#requests
    .. _tqdm:
        https://pypi.org/project/tqdm/

    """
    from ._internal_support._local_or_remote import get_local_or_remote as _impl

    return _impl(
        file=_paths.any_path_to_str(file),
        local_root=_paths.any_path_to_str(local_root),
        remote_root=_paths.any_path_to_str(remote_root),
        force=force,
        postprocessor=postprocessor,
        show_progress=show_progress,
    )


def serializable(obj: object) -> bool:
    """Check if `obj` is serializable using Pickle.

    Args:
        obj: Object to test.

    Returns:
        ``True`` if `obj` was pickled without issues.

    """
    import io
    import pickle

    bio = io.BytesIO()
    try:
        pickle.dump(obj, bio)
        return True  # noqa: TRY300
    except Exception:
        return False



import os
import sys
from datetime import UTC, datetime
from getpass import getuser
from typing import Any, Literal, NotRequired, TypedDict

from rics import __version__


class KernelSpec(TypedDict):
    """Simplified `kernel.json` type. See the `Jupyter kernel documentation`_ for details.

    .. _Jupyter kernel documentation: https://jupyter-client.readthedocs.io/en/stable/kernels.html#kernel-specs
    """

    argv: list[str]
    display_name: str
    language: Literal["python"]
    metadata: dict[str, Any]
    env: NotRequired[dict[str, str]]


def set_metadata(metadata: dict[str, Any]) -> None:
    """Set kernelspec metadata."""
    metadata["created_at"] = datetime.now(UTC).replace(microsecond=0).isoformat()
    metadata["created_by"] = getuser()

    metadata["installer"] = {
        "version": __version__,
        "sys.executable": sys.executable,
        "sys.version": sys.version,
        "workdir": os.getcwd(),  # noqa: PTH109
    }



"""Action level enumeration types."""

import typing as _t
from enum import Enum as _Enum


class ActionLevel(_Enum):
    """Action level enumeration type for events."""

    _ignore_: _t.ClassVar = ("ParseType",)  # type: ignore[assignment, misc]

    ParseType: _t.TypeAlias = "str | ActionLevel"
    """Types that may be interpreted as an ``ActionLevel``."""

    RAISE = "raise"
    """Raise an error."""
    WARN = "warn"
    """Raise a warning"""
    IGNORE = "ignore"
    """Ignore the event."""

    @classmethod
    def verify(
        cls,
        action: ParseType,
        purpose: str | None = None,
        forbidden: "ActionLevel.ParseType | None" = None,
    ) -> "ActionLevel":
        """Verify an action level.

        Args:
            action: The value to verify.
            purpose: Additional information to add if an exception is raised.
            forbidden: Which of ('ignore', 'warn', 'raise') not to accept. Accept all if ``None``.

        Returns:
            A valid action level for the given purpose.

        Raises:
            BadActionLevelError: If `action` is not in ('ignore', 'warn', 'raise') or if `action` is in `forbidden`.

        """
        forbidden = None if forbidden is None else cls.verify(forbidden)

        if isinstance(action, ActionLevel):
            action_level = action
        else:
            try:
                action_level = ActionLevel[action.upper()]
            except KeyError:
                raise BadActionLevelError(action, purpose, forbidden) from None
            except AttributeError:
                raise BadActionLevelError(action, purpose, forbidden) from None

        if action_level is forbidden:
            raise BadActionLevelError(action, purpose, forbidden)

        return action_level

    def __eq__(self, other: _t.Any) -> bool:
        if isinstance(other, ActionLevel):
            other = other.name

        if isinstance(other, str):
            return bool(self.name == other.upper())

        return False

    def __hash__(self) -> int:
        return hash(self.name)


ActionLevel.ParseType = _t.Literal["ignore", "warn", "raise", "IGNORE", "WARN", "RAISE"] | ActionLevel  # type: ignore


class ActionLevelHelper:
    """Helper class for keeping track of per-purpose forbidden actions.

    Args:
        require_purpose: Determine how to handle the `purpose` argument given in :meth:`verify`; require that it exists,
            or just that it is given. Must be one of ('given', 'exists').
        **forbidden_for_purpose: Per-purpose forbidden actions.
    """

    def __init__(
        self,
        require_purpose: _t.Literal["given", "exists"] = "exists",
        **forbidden_for_purpose: "ActionLevel.ParseType | None",
    ) -> None:
        self._require = require_purpose
        self._forbidden_for_purpose = forbidden_for_purpose

    def verify(self, action: ActionLevel.ParseType, purpose: str) -> ActionLevel:
        """Verify an action level with constraints.

        Args:
            action: The value to verify.
            purpose: Additional information to add if an exception is raised.

        Returns:
            A valid action level for the given purpose.

        Raises:
            BadActionLevelError: If `action` is not in ('ignore', 'warn', 'raise') or if `action` is in `remove`.
            ValueError: If `purpose` is unknown and `require_purpose` is set to `'exists'`.

        """
        if self._require == "exists" and purpose not in self._forbidden_for_purpose:
            raise ValueError(f"Unknown {purpose=} given.")
        return ActionLevel.verify(action, purpose, forbidden=self._forbidden_for_purpose.get(purpose))


class BadActionLevelError(ValueError):
    """Error raised for unknown or disallowed action arguments."""

    def __init__(
        self,
        action: ActionLevel.ParseType,
        purpose: str | None = None,
        forbidden: "ActionLevel | None" = None,
    ) -> None:
        extra = f" for {purpose}" if purpose else ""
        all_options = {ActionLevel.RAISE, ActionLevel.WARN, ActionLevel.IGNORE}
        if forbidden:
            all_options.discard(forbidden)
        accepted = tuple(a.name.lower() for a in all_options)
        super().__init__(f"Permitted choices{extra} are {accepted}, but got {action=}.")



"""Read environment variables as specific types.

To read ``typing.Literal`` values, use :meth:`rics.types.LiteralHelper.read_env` instead.
"""

from ._base import read_env
from ._bool import read_bool
from ._enum import read_enum
from ._numeric import read_float, read_int
from ._str import read_str

__all__ = [
    "read_bool",
    "read_enum",
    "read_env",
    "read_float",
    "read_int",
    "read_str",
]



"""CLI entrypoint.

For details, see the :doc:`CLI documentation </documentation/cli/notebooks/cli>` page.
"""

from importlib.metadata import EntryPoint as _EntryPoint
from importlib.metadata import entry_points as _entry_points

import click as _c

from rics import click as _rc
from rics import logs as _logs
from rics.click import _alias_command_group


@_c.group(
    context_settings={"auto_envvar_prefix": "RICS"},
    epilog="""
    \b
    Python package documentation:
        https://id-translation.readthedocs.io/
        https://time-split.readthedocs.io/
        https://rics.readthedocs.io/
    Please let me know if any of these resources helped you!
    """,
    cls=_alias_command_group.AliasedCommandGroup,
)
@_rc.logging_verbosity_option(
    "-v",
    "--verbose",
    mode="forward_both",
    format=_logs.FORMAT_MS,
    levels=[{"rics": "INFO"}, {"root": "INFO"}, {"rics": "DEBUG"}, {"root": "DEBUG"}],
    show_envvar=True,
)
@_c.pass_context
def main(cxt: _c.Context, verbose: tuple[int, _logs.LoggingSetupHelper]) -> None:
    """👻 RiCS: My little CLI program."""
    _rc.get_user_command(cxt, meta=True)

    # Configure logging "manually" since mode="forward_both" does not call LoggingSetupHelper.configure_logging().
    verbosity, helper = verbose
    helper.configure_logging(verbosity)
    cxt.meta["verbosity"] = verbosity
    cxt.meta["logging_setup_helper"] = helper


def _register_dummy(entrypoint: _EntryPoint, exc: Exception) -> None:  # pragma: no cover
    import sys

    @_c.command(entrypoint.name, help=f"[unavailable] {exc}", add_help_option=False)
    def _dummy(*kwargs, **args):  # type: ignore # noqa
        msg = f"Command '{entrypoint.name}' @ '{entrypoint.value}' is unavailable: "

        _c.secho(msg, err=True, fg="white", nl=False)
        _c.secho(repr(exc), err=True, fg="red")
        sys.exit(1)

    main.add_command(_dummy, entrypoint.name)


def _register_entrypoints() -> None:
    for ep in _entry_points(group="rics.cli"):
        try:
            result = ep.load()

            if isinstance(result, _c.Command):
                main.add_command(result, ep.name)
            else:
                msg = f"not a {_c.Command.__module__}.{_c.Command.__qualname__} subtype"
                _register_dummy(ep, TypeError(msg))
        except Exception as e:
            _register_dummy(ep, e)


_register_entrypoints()

if __name__ == "__main__":
    main()



"""Utility for splitting a ``CHANGELOG.md`` file into multiple pages.

Not mean for general consumption; this is used to split the ``rics`` changelog, and makes several assumptions about the
format which do not hold in general. Included here because it makes my life easier.
"""

from ._patch_notes import PatchNotes
from ._split_changelog import split_changelog
from ._write_changelog import write_changelog

__all__ = ["PatchNotes", "split_changelog", "write_changelog"]



"""Environment variable interpolation in files and strings.

This module provides utilities to read text and replace environment variable references with the actual environment
variable value using a familiar syntax. Optional default values are also possible, as well as recursively nested
variables (must be explicitly enabled).

Syntax:
    Similar to Bash variable interpolation; ``${<var>}``, where ``<var>`` is the name of the environment variable you
    wish to extract. This particular form will raise an exception if ``<var>`` is not set.

Default values:
    Default values are specified by stating the default value after a colon; ``${<var>:default-value}``. The default
    value may be blank, in which case an empty string is used as the default.

There are three main ways of using this module:

    * The :meth:`.Variable.parse_string`-function, combined with :meth:`.Variable.get_value`. This gives you the most
      amount of control.
    * The :func:`replace_in_string`-function, for basic interpolation without recursion in an entire string.
    * The :func:`rics.misc.interpolate_environment_variables`-function, which adds some additional logic on top of what
      the :class:`Variable`-methods provide.

Examples:
    We'll set ``ENV_VAR0=VALUE0`` and ``ENV_VAR1=VALUE1``. Vars 2 and 4 are not set.

    >>> from os import environ
    >>> environ["ENV_VAR0"] = "VALUE0"
    >>> environ["ENV_VAR1"] = "VALUE1"

    Basic interpolation.

    >>> Variable.parse_first("${ENV_VAR0}").get_value()
    'VALUE0'

    Setting a default works as expected. The default may also be empty.

    >>> Variable.parse_first("${ENV_VAR2:default}").get_value()
    'default'
    >>> Variable.parse_first("${ENV_VAR2:}").get_value()
    ''

    The variable must exist if no default is given.

    >>> Variable.parse_first("${DOESNT_EXIST}").get_value()  # doctest: +SKIP
    UnsetVariableError: Required Environment Variable 'DOESNT_EXIST': Not set.

    Furthermore, variables may be nested, but this requires setting the flag to enable recursive parsing.

    >>> Variable.parse_first("${ ENV_VAR2 :${ ENV_VAR0 }}").get_value(
    ...     resolve_nested_defaults=True
    ... )
    'VALUE0'

    Whitespaces around the names is fine. Finally, nesting may be arbitrarily deep.

    >>> Variable.parse_first("${ENV_VAR3:${ENV_VAR4:${ENV_VAR0}}}").get_value(True)
    'VALUE0'
"""

from ._file_utils import replace_in_string
from ._variable import UnsetVariableError, Variable

__all__ = ["UnsetVariableError", "Variable", "replace_in_string"]



"""Utility methods for working with paths and filesystems."""

import typing as _t
from pathlib import Path as _Path

from rics.types import AnyPath as _AnyPath

AnyPath = _AnyPath
"""Reexport of :attr:`rics.types.AnyPath`."""
PathT = _t.TypeVar("PathT", str, _Path)  # Would be nice to allow Any or a bound
"""A specific path-like type."""
PathPostprocessor = _t.Callable[[PathT], PathT | bool | None]


def any_path_to_str(
    path: _AnyPath,
    *,
    postprocessor: PathPostprocessor[str] | None = None,
) -> str:
    """Docstring is generated."""
    return parse_any_path(path, cls=str, postprocessor=postprocessor)


def any_path_to_path(
    path: _AnyPath,
    *,
    postprocessor: PathPostprocessor[_Path] | None = None,
) -> _Path:
    """Docstring is generated."""
    return parse_any_path(path, cls=_Path, postprocessor=postprocessor)


def parse_any_path(
    path: _AnyPath,
    *,
    cls: type[PathT],
    postprocessor: PathPostprocessor[PathT] | None = None,
) -> PathT:
    """Docstring is generated."""
    converted_path = _convert_path(path, cls=cls)

    if postprocessor is not None:
        result = postprocessor(converted_path)
        if isinstance(result, cls):
            converted_path = result
        elif result is None or result is True:
            pass  # postprocessor was a validation function
        elif result is False:
            from rics.misc import tname

            msg = f"Validator {tname(postprocessor, prefix_classname=True)}({converted_path!r}) returned False."
            raise ValueError(msg)
        else:
            from rics.misc import tname

            pretty_actual = f"`{tname(result)}`"
            pretty_expected = f"({tname(cls)} | bool | None)"
            pretty_pp = f"{tname(postprocessor)}({converted_path!r})"
            msg = f"Postprocessor {pretty_pp} returned {result=} of type {pretty_actual}; expected={pretty_expected}."
            raise TypeError(msg)

    return converted_path


def _convert_path(path: _AnyPath, *, cls: type[PathT]) -> PathT:
    if isinstance(path, cls):
        return path

    if issubclass(cls, str):
        return cls(path)

    if issubclass(cls, _Path) and isinstance(path, str):
        return cls(path)

    try:
        return cls(str(path))
    except Exception as e:
        msg = f"Cannot convert {path=} from type(path)=`{type(path).__name__}` to cls=`{cls.__name__}`."
        raise TypeError(msg) from e


def _set_path_func_docstring(func: _t.Any, cls: str | type[_t.Any], tail: str = "") -> None:
    args = {"path": "Any :class:`path-like <rics.paths.AnyPath>` type."}

    if isinstance(cls, type):  # Specialization
        cls = cls.__name__

        equivalent_to = f"""Equivalent to :func:`parse_any_path(path, cls={cls}) <.parse_any_path>`."""
        args["postprocessor"] = f"A callable ``({cls}) -> {cls} | None``."

        cls = f"``{cls}``"
    else:  # Base function
        equivalent_to = ""
        args["cls"] = "Desired output type; typically one of :py:class:`str` | :py:class:`pathlib.Path`."
        args["postprocessor"] = "A callable ``(PathT) -> PathT | bool | None``."

    args["postprocessor"] += " Returns the post-processed value if `postprocessor` does not return ``None | bool``."

    args_str = "\n        ".join(f"{arg}: {docstring}" for arg, docstring in args.items())
    docstring = f"""Convert a path of any type to {cls}.

    {equivalent_to}

    Args:
        {args_str}

    Returns:
        A `path` of type {cls}.

    {tail}
    """

    func.__doc__ = docstring


if __doc__:
    _set_path_func_docstring(
        parse_any_path,
        cls="of type `cls`",
        tail="""Raises:
        TypeError: If the `postprocessor` returns an invalid type.
        ValueError: If the `postprocessor` returns ``False``.

    Postprocessor return types:
        * ``None``: Validator. Should raise if there's an issue.
        * ``bool``: Validator, e.g. :meth:`~pathlib.Path.is_dir`. Raises a ``ValueError`` if the
          function returns ``False``.
        * ``PathT``: Postprocessor, e.g. :meth:`~pathlib.Path.absolute`. Result is returned as-is.

    Examples:
        >>> from pathlib import Path
        >>> parse_any_path("/home/dev/", cls=Path)
        PosixPath('/home/dev')

        >>> parse_any_path(Path("/home/dev/"), cls=Path)
        PosixPath('/home/dev')

    See Also:
        * :func:`any_path_to_str`: Alias of ``parse_any_path(cls=str)``
        * :func:`any_path_to_path`: Alias of ``parse_any_path(cls=pathlib.Path)``
""",
    )
    _set_path_func_docstring(any_path_to_str, cls=str)
    _set_path_func_docstring(any_path_to_path, cls=_Path)



import datetime
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Self


@dataclass(frozen=True)
class PatchNotes:
    """Notes for a single release."""

    title: str
    """Title of the release, such as `'0.1.0'`."""
    date: str
    """Date of the release."""
    extras: list[str]
    """Extra content to append just below the title."""

    sections: dict[str, list[str]]
    """Subsections such as `'Added'` and `'Fixed'`."""

    def __iter__(self) -> Iterable[tuple[str, list[str]]]:
        for section in sorted(self.sections):
            lines = self.sections[section]
            if lines:
                yield section.title(), lines

    @classmethod
    def from_markdown(cls, raw: str) -> tuple[list[Self], dict[str, str]]:
        """Create ``PatchNotes`` and references from the lines of a Markdown file.

        Args:
            raw: Raw blob of ``CHANGELOG.md`` file.

        Returns:
            A tuple ``([patch_notes..], {version: github_link})``.

        Raises:
            ValueError: If there are duplicate subsections.

        """
        parts = raw.split("\n## ")[1:]  # First section is just the title itself.
        last_lines = parts[-1].splitlines()  # Last part has references that we need to extract.
        del parts[-1]

        parsed_notes = list(map(cls._consume, parts))

        last_section_lines: list[str] = []
        for references_start, line in enumerate(last_lines):
            if "]: http" in line:
                parsed_notes.append(cls._consume("\n".join(last_section_lines)))
                refs = {line.partition(": ")[0].strip(): line for line in last_lines[references_start:]}
                return parsed_notes, refs
            else:
                last_section_lines.append(line)

        raise ValueError("Failed to parse final section:\n", "\n".join(last_lines))

    @classmethod
    def _consume(cls, content: str) -> Self:
        title, _, the_rest = map(str.strip, content.partition("\n"))

        if "-" in title:
            title, _, date = title.partition("-")
            date = datetime.datetime.strptime(date.strip(), "%Y-%m-%d").strftime("%B %d, %Y")
        else:
            date = ""

        subsection_content_lines: list[list[str]] = list(map(str.splitlines, the_rest.split("### ")))
        extra = subsection_content_lines[0]

        sections = {section_lines[0]: section_lines[1:] for section_lines in subsection_content_lines[1:]}

        return cls(title.strip(), date.strip(), extra, sections)



"""Types used by the plotting framework."""

import typing as _t
from collections.abc import Callable as _Callable

if _t.TYPE_CHECKING:
    try:
        from seaborn import FacetGrid
    except ImportError:
        FacetGrid = _t.Never

Candidate = _t.Literal["Candidate"]
TestData = _t.Literal["Test data"]
FuncOrData = Candidate | TestData
"""Valid axis preferences."""

Kind = _t.Literal["bar", "box", "boxen", "point", "strip", "swarm", "violin"]
"""Valid plot kinds."""
Unit = _t.Literal["s", "ms", "μs", "us", "ns"]
"""Valid time units."""

Postprocessor = _Callable[["FacetGrid"], None]
"""A callable ``(FacetGrid) -> None`` which applies fixups to a :class:`seaborn.FacetGrid`."""



from collections.abc import Hashable, Iterable, Mapping
from dataclasses import dataclass, field, fields
from typing import Any, ClassVar, Literal, Self

import numpy as np
import pandas as pd

from rics.collections.dicts import compute_if_absent
from rics.performance.plot_types import Candidate, FuncOrData, Kind, TestData, Unit
from rics.types import LiteralHelper

from ..types import ResultsDict

FUNC: Candidate = "Candidate"
DATA: TestData = "Test data"


@dataclass(frozen=True, kw_only=True)
class CatplotParams:
    data: pd.DataFrame = field(repr=False)
    x: FuncOrData
    y: str
    hue: FuncOrData
    kind: Kind

    horizontal: bool = field(metadata={"skip": True})
    names: list[str] = field(metadata={"skip": True})
    user_kwargs: Mapping[str, Any] = field(metadata={"skip": True})

    DEFAULTS: ClassVar[dict[str, Any]] = {"errorbar": "sd", "estimator": "min", "aspect": 2}
    DEFAULTS_BY_KIND: ClassVar[dict[Kind, dict[str, Any]]] = {
        "bar": {"capsize": 0.2},
        "point": {"capsize": 0.2},
    }
    COMPUTED_DEFAULTS: ClassVar[tuple[str, ...]] = ("order", "hue_order", "log_scale")

    @classmethod
    def reserved_keys(cls) -> set[str]:
        return {f.name for f in fields(cls) if not f.metadata.get("skip")}

    def __post_init__(self) -> None:
        if keys := self.reserved_keys().intersection(self.user_kwargs):
            msg = f"Bad `kwargs`: {keys=} are reserved."
            error = ValueError(msg)

            if "y" in keys:
                y = self.user_kwargs["y"]
                error.add_note(f"HINT ({y=}): Use `horizontal=True` to plot timings on the X-axis.")
            if "hue" in keys:
                hue = self.user_kwargs["hue"]
                error.add_note(f"HINT ({hue=}): Use `x='data' | 'candidate'` to control X-axis and Hue variables.")

            raise error

        helper = LiteralHelper[FuncOrData](FuncOrData)
        helper.check(self.x, "x")
        helper.check(self.hue, "hue")

    @classmethod
    def make(
        cls,
        run_results: ResultsDict | pd.DataFrame,
        *,
        x: Literal["candidate", "data"] | None = None,
        horizontal: bool = False,
        unit: Unit | None = None,
        kind: Kind = "bar",
        names: Iterable[str] = (),
        **kwargs: Any,
    ) -> Self:
        """Create instance from run results."""
        names = [*names]
        df = _make_df(run_results, names)

        # MyPy thinks these are both string
        x_col, hue_col = (DATA, FUNC) if _is_data_x(x, df=df) else (FUNC, DATA)

        return cls(
            data=df,
            x=x_col,
            y=_resolve_y(unit, df=df),
            horizontal=horizontal,
            hue=hue_col,
            kind=kind,
            names=names,
            user_kwargs={**kwargs},
        )

    def to_kwargs(self) -> dict[str, Any]:
        """Convert to :func:`seaborn.catplot` keyword arguments."""
        kwargs = dict(self.user_kwargs)

        for key in self.reserved_keys():
            kwargs[key] = getattr(self, key)

        for key, default in self.DEFAULTS.items():
            kwargs.setdefault(key, default)

        for key, default in self.DEFAULTS_BY_KIND.get(self.kind, {}).items():
            kwargs.setdefault(key, default)

        for key in self.COMPUTED_DEFAULTS:
            compute_if_absent(kwargs, key, self._compute_default)

        if self.names:
            self._handle_row_col(kwargs)

        if self.horizontal:
            x = kwargs["x"]
            kwargs["x"] = kwargs["y"]
            kwargs["y"] = x

        return kwargs

    @property
    def want_log_scale_hack(self) -> bool:
        return self.kind == "bar"

        user_log_scale = self.user_kwargs.get("log_scale")
        if not user_log_scale:
            return True  # No explicit user choice - apply

        # TODO(me): bool, number, or pair of bools or numbers

        # and isinstance(self.user_kwargs.get("log_scale"), tuple)

    def _compute_default(self, key: str) -> Any:
        df = self.data

        match key:
            case "order":
                return df[self.x].unique().tolist()
            case "hue_order":
                return df[self.hue].unique().tolist()
            case "log_scale":
                column = next(filter(lambda c: c.startswith("Time ["), df.columns))
                means = df.groupby([DATA, FUNC], observed=True)[column].mean()
                return (False, True) if means.max() / means.min() > 20 else None  # noqa: PLR2004

        raise KeyError(f"Bad {key=}. Possible choices={self.COMPUTED_DEFAULTS}.")

    def _handle_row_col(self, kwargs: dict[str, Any]) -> None:
        names = self.names
        updates = {}

        # We could print something like f"{label}: [row]", but this could add a lot of additional text to the legend
        # or x-axis. I think this is a better way to do it.
        hidden = {names.index(name) for key in ("row", "col") if (name := kwargs.get(key)) in names}

        # Would be nice if data label formatting was configurable. For now though, we'll settle for this.
        formatters = {i: f"{name} = {{}}" for i, name in enumerate(names) if i not in hidden}

        def format_label(label: tuple[Hashable]) -> str:
            if not isinstance(label, tuple) or len(label) != len(names):
                raise TypeError(f"Cannot format {label=} using {names=}.")
            g = (template.format(lv) for i, lv in enumerate(label) if (template := formatters.get(i)))
            return " | ".join(g)

        order_key = "hue_order" if self.hue == DATA else "order"
        if old_order := kwargs.get(order_key):
            updates[order_key] = np.unique([format_label(label) for label in old_order]).tolist()

        # assert sorted(self.data.columns.intersection(names)) == sorted(names)

        data = kwargs["data"].copy()
        data[DATA] = data[DATA].map(format_label)
        updates["data"] = data

        # Don't write updates until we're done; allows caller to skip if needed.
        kwargs.update(updates)


def _make_df(run_results: ResultsDict | pd.DataFrame, names: list[str]) -> pd.DataFrame:
    from rics.performance import to_dataframe

    if isinstance(run_results, dict):
        df = to_dataframe(run_results, names=names)
    else:
        df = run_results.copy()

        if names and {*names}.difference(df.columns):
            df[names] = pd.DataFrame.from_records(df[DATA])

        return df

    as_category = [DATA, FUNC]
    df[as_category] = df[as_category].astype("category")
    return df


def _is_data_x(x_col: Literal["candidate", "data"] | None, *, df: pd.DataFrame) -> bool:
    if x_col is None:
        n_data, n_func = df[[DATA, FUNC]].nunique()
        return n_data > n_func  # type: ignore[no-any-return]
    else:
        return x_col.lower().startswith("d")


def _resolve_y(unit: Unit | None, *, df: pd.DataFrame) -> str:
    if unit is None:
        return _compute_nice_y(df)

    if unit == "us":
        unit = "μs"

    y = f"Time [{unit}]"
    if y not in df:
        msg = f"Bad {unit=}; column '{y}' not present in data."
        raise TypeError(msg)
    return y


def _compute_nice_y(df: pd.DataFrame) -> str:
    """Pick the unit with the most "human" scale; whole numbers around one hundred."""
    from numpy import log10

    columns = [c for c in df.columns if c.startswith("Time [")]
    means = df.groupby([DATA, FUNC], observed=True)[columns].mean()

    residuals = log10(means) - 2
    avg_residual_by_time_column = residuals.mean(axis="index")
    y = avg_residual_by_time_column.abs().idxmin()

    assert isinstance(y, str)  # noqa: S101
    return y



from ._variable import UnsetVariableError, Variable


def replace_in_string(
    s: str,
    *,
    allow_nested: bool = True,
    allow_blank: bool = False,
) -> str:
    """Interpolate environment variables in a string `s`.

    This function replaces references to environment variables with the actual value of the variable, or a default if
    specified. The syntax is similar to Bash string interpolation; use ``${<var>}`` for mandatory variables, and
    ``${<var>:default}`` for optional variables.

    Args:
        s: A string in which to interpolate.
        allow_blank: If ``True``, allow variables to be set but empty.
        allow_nested: If ``True`` allow using another environment variable as the default value. This option will not
            verify whether the actual values are interpolation-strings.

    Returns:
        A copy of `s`, after environment variable interpolation.

    Raises:
        ValueError: If nested variables are discovered (only when ``allow_nested=False``).
        UnsetVariableError: If any required environment variables are unset or blank (only when ``allow_blank=False``).
    """
    for var in Variable.parse_string(s):
        if not allow_nested and (var.default and Variable.parse_string(var.default)):
            raise ValueError(f"Nested variables forbidden since {allow_nested=}.")

        value = var.get_value(resolve_nested_defaults=allow_nested).strip()

        if not (allow_blank or value):
            msg = f"Empty values forbidden since {allow_blank=}."
            raise UnsetVariableError(var.name, msg)

        s = s.replace(var.full_match, value)
    return s



"""Plotting utility methods."""

import typing as _t

from matplotlib.axis import Axis as _Axis
from matplotlib.axis import XAxis as _XAxis
from matplotlib.axis import YAxis as _YAxis
from matplotlib.ticker import FuncFormatter as _FuncFormatter
from matplotlib.ticker import IndexLocator as _IndexLocator

ERROR_BAR_CAPSIZE: float = 0.1

HalfRep = _t.Literal["fraction", "f", "decimal", "d"]
Decimals = int | _t.Literal[False] | None


class HasXAxis(_t.Protocol):
    """Protocol class indicating something that as an X-axis."""

    xaxis: _XAxis
    """X-Axis attribute."""


class HasYAxis(_t.Protocol):
    """Protocol class indicating something that as an Y-axis."""

    yaxis: _YAxis
    """Y-Axis attribute."""


def configure() -> None:
    """Call all configure-functions in this module."""
    configure_matplotlib()
    configure_seaborn()


def configure_seaborn() -> None:
    """Configure Seaborn figure plotting.

    Caveat Emptor: May do strange stuff 👻.

    Raises:
        ModuleNotFoundError: If Seaborn is not installed.

    """
    import functools
    import warnings

    import seaborn as sns

    warnings.filterwarnings("ignore", module="seaborn")

    sns.set_theme(context="talk")

    sns.barplot = functools.partial(sns.barplot, capsize=ERROR_BAR_CAPSIZE)
    # Doesn't play nice with all plot kinds
    # sns.catplot = functools.partial(sns.catplot, capsize=ERROR_BAR_CAPSIZE, height=5)


def configure_matplotlib() -> None:
    """Configure Matplotlib figure plotting.

    Caveat Emptor: May do strange stuff 👻.

    Raises:
        ModuleNotFoundError: If matplotlib is not installed.

    """
    import matplotlib.pyplot as plt

    plt.rcParams["figure.figsize"] = (20, 5)
    plt.rcParams["figure.autolayout"] = True


def pi_ticks(ax: _Axis | HasXAxis, half_rep: HalfRep | None = None) -> None:
    """Decorate an axis by setting the labels to multiples of pi.

    .. image:: ../_images/pi_ticks.png

    .. list-table:: Options for the `half_rep` argument.
       :header-rows: 1

       * - Value
         - Interpretation
         - Example output
       * - ``None``
         - Show integer multiples only.
         - `0, π, 2π, 3π, ...`
       * - `'f'` or `'fraction'`
         - Halves of `π` use fractional representation.
         - `0/2π, 1/2π, 2/2π, 3/2π, ...`
       * - `'d'` or `'decimal'`
         - Halves of `π` use decimal representation.
         - `0.0π, 0.5π, 1.0π, 1.5π, ...`

    Args:
        ax: An axis to decorate, or an object with an `xaxis` attribute.
        half_rep: Controls how fractions of `π` are represented on the x-axis.

    """
    axis: _Axis = ax.xaxis if hasattr(ax, "xaxis") else ax

    helper = _PiTickHelper(half_rep, start=axis.get_data_interval()[0])
    axis.set_major_locator(helper.locator)
    axis.set_major_formatter(helper.formatter)


class _PiTickHelper:
    PI: float = 3.14159265359
    HALF_PI: float = PI / 2

    def __init__(
        self,
        half_rep: HalfRep | None,
        *,
        start: float,
    ) -> None:
        self.half = self._parse_half_rep(half_rep)
        r = self.HALF_PI if half_rep else self.PI
        self.offset = r * round(start / r) - start  # Offset from nearest (half) multiple of pi.

        self.locator = _IndexLocator(base=self.HALF_PI if self.half else self.PI, offset=self.offset)
        self.formatter = _FuncFormatter(self._format)

    @staticmethod
    def _parse_half_rep(
        half_rep: str | None,
    ) -> _t.Literal["frac", "dec"] | None:
        if half_rep is None:
            return None

        parsed_half = half_rep.lower()
        if parsed_half.startswith("f"):
            return "frac"
        if parsed_half.startswith("d"):
            return "dec"

        msg = f"Argument {half_rep=} not in ('[f]raction', '[d]ecimal', None)."
        raise TypeError(msg)

    def _format(self, x: float, _pos: int) -> str:
        n = round(x / self.PI, 1)

        if self.half is None:
            if x == 0:
                return "0"
            if x == self.PI:
                return "π"

            return f"{int(n)}π"

        if self.half == "dec":
            return f"{n}π"

        return f"{int(n * 2)}/2π"


def percentage_ticks(
    ax: _Axis | HasYAxis,
    *,
    sign: bool = False,
    decimals: Decimals = 1,
) -> None:
    """Decorate an axis by formatting ticks as percentages.

    Args:
        ax: An axis to decorate, or an object with a `yaxis` attribute.
        sign: If ``True``, show prepend `'+'` for positive ticks.
        decimals: Number of decimals to keep.

    Return:
        The formatting string.
    """
    axis: _Axis = ax.yaxis if hasattr(ax, "yaxis") else ax

    formatter = _make_percent_formatter(sign, decimals)
    axis.set_major_formatter(formatter)


def _make_percent_formatter(sign: bool, decimals: Decimals) -> str:
    plus = "+" if sign else ""
    decimals = decimals or 0
    return "{" + f"x:{plus}.{decimals}%" + "}"



"""Utility functions and classes for working with https://pypi.org/project/click/ CLI applications."""

from ._logging import logging_verbosity_option
from ._util import USER_COMMAND_KEY, get_user_command

__all__ = [
    "USER_COMMAND_KEY",
    "get_user_command",
    "logging_verbosity_option",
]



from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from rics.types import EnumT


@overload
def read_enum(var: str, default_or_type: "EnumT | type[EnumT]", *, split: str) -> list["EnumT"]: ...
@overload
def read_enum(var: str, default_or_type: "EnumT | type[EnumT]", *, split: None = None) -> "EnumT": ...
def read_enum(var: str, default_or_type: "EnumT | type[EnumT]", *, split: str | None = None) -> "EnumT | list[EnumT]":
    """Read ``enum`` variable.

    This function wraps :meth:`.LiteralHelper.from_enum` and :meth:`.LiteralHelper.read_env`.

    Args:
        var: Variable to read.
        default_or_type: Enum type or default enum value. Must be a concrete value (e.g. ``Animal.dog`` instead of just
            ``Animal``) to allow unset or blank `var` values.
        split: Character to split on. Returns ``list[EnumT]`` when set.

    Returns:
        An :attr:`~rics.types.EnumT` value, or a list thereof (if `split` is set).

    Raises:
        TypeError: If conversion fails.

    Examples:
        Basic usage.

        >>> from enum import Enum
        >>> class Animal(Enum):
        ...     cat = "meow"
        ...     dog = "woof"

        Enums are matched by name, never by value.

        >>> import os
        >>> os.environ["MY_ANIMAL"] = "cat"
        >>> read_enum("MY_ANIMAL", Animal)
        <Animal.cat: 'meow'>

        Input is always cleaned and normalized.

        >>> os.environ["MY_ANIMAL"] = " DOG"
        >>> read_enum("MY_ANIMAL", Animal)
        <Animal.dog: 'woof'>

        This also applies when reading multiple enums.

        >>> os.environ["MY_ANIMAL_LIST"] = " DOG, cat  , CAT"
        >>> read_enum("MY_ANIMAL_LIST", Animal, split=",")
        [<Animal.dog: 'woof'>, <Animal.cat: 'meow'>, <Animal.cat: 'meow'>]


        Unlike the other functions in this module, ``read_enum`` always runs in `strict` mode. Define a default for when
        the `var` is not set or blank by passing concrete value in the second position.

        >>> read_enum("MISSING_VARIABLE", Animal.dog)
        <Animal.dog: 'woof'>

        Unknown values will always raise.
    """
    from rics.types import LiteralHelper

    if isinstance(default_or_type, type):
        cls = default_or_type
        default = None
    else:
        cls = type(default_or_type)
        default = default_or_type

    return LiteralHelper.from_enum(cls).read_env(var, default=default, split=split)



from functools import partial

from docutils.nodes import Node
from sphinx.addnodes import toctree
from sphinx.ext.autosummary import Autosummary, autosummary_toc

_real = Autosummary.run


def patch(*, max_parents: int = 1) -> None:
    """Use shorter TOC tree entries.

    For example, with `max_parents=1`, a TOC title `'rics.collections.dicts'` will be replaced by `'collections.dicts'`.

    Args:
        max_parents: Maximum number of parent modules to show.
    """
    if max_parents < 0:
        raise ValueError(f"bad {max_parents=} < 0")

    if Autosummary.run is _real:
        Autosummary.run = partial(_hack, n=max_parents + 1)  #  type: ignore[method-assign]
        print("patch: Autosummary.run")


def _hack(self: Autosummary, n: int) -> list[Node]:
    nodes = _real(self)

    for node in nodes:
        if not isinstance(node, autosummary_toc):
            continue

        for child in node.children:
            if not isinstance(child, toctree):
                continue

            entries = child["entries"]

            for i, (original_title, ref) in enumerate(entries):
                if original_title is None and ref.count(".") >= n:
                    new_title = ".".join(ref.rsplit(".", n)[-n:])
                    entries[i] = (new_title, ref)

    return nodes



import click


class AliasedCommandGroup(click.Group):
    """Work in progress."""

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if exact := super().get_command(ctx, cmd_name):
            return exact

        candidates = self.list_commands(ctx)

        for matcher in self._startswith, self._dashes:
            matches = matcher(candidates, cmd_name)
            if len(matches) == 1:
                match = matches[0]
                # print(f"{type(self).__name__}: Subcommand {match!r} selected by matcher={matcher.__name__}().")
                return super().get_command(ctx, match)

        # ctx.fail(f"Ambiguous alias `{cmd_name}`: {', '.join(sorted(matches))}")  # Multiple matches
        return None

    @classmethod
    def _startswith(cls, candidates: list[str], cmd_name: str) -> list[str]:
        return [c for c in candidates if c.startswith(cmd_name)]

    @classmethod
    def _dashes(cls, candidates: list[str], cmd_name: str) -> list[str]:
        return [c for c in candidates if "".join(part[0] for part in c.split("-")) == cmd_name]

    def resolve_command(
        self, ctx: click.Context, args: list[str]
    ) -> tuple[str | None, click.Command | None, list[str]]:
        _, cmd, args = super().resolve_command(ctx, args)

        if cmd is None:
            return None, None, args

        return cmd.name, cmd, args



from typing import overload

from ._base import read_env


@overload
def read_int(var: str, default: int = 0, *, strict: bool = True, split: str) -> list[int]: ...
@overload
def read_int(var: str, default: int = 0, *, strict: bool = True, split: None = None) -> int: ...
def read_int(var: str, default: int = 0, *, strict: bool = True, split: str | None = None) -> int | list[int]:
    """Read ``int`` variable.

    Args:
        var: Variable to read.
        default: Default value to use when `var` is not set (or blank).
        strict: If ``False``, always fall back to `default` instead of raising.
        split: Character to split on. Returns ``list[int]`` when set.

    Returns:
        An ``int`` value, or a list thereof (if `split` is set).

    Examples:
        Basic usage.

        >>> import os
        >>> os.environ["MY_INT"] = "1"
        >>> read_int("MY_INT")
        1

        When using ``strict=False``, unmapped values are converted to the `default` instead of raising.

        >>> os.environ["MY_INT"] = "not-an-int"
        >>> read_int("MY_INT", strict=False)
        0

        Note that this includes passing ``float`` values, even if the decimal is zero.

        >>> os.environ["MY_INT"] = "0.0"
        >>> read_int("MY_INT", default=2019, strict=False)
        2019

        When using `split`, elements are cleaned individually and blank items are skipped.

        >>> os.environ["MY_INT_LIST"] = "0, 1, , 5"
        >>> read_int("MY_INT_LIST", split=",")
        [0, 1, 5]

        Conversion must succeed for all elements, or the `default` will be returned.
    """
    return read_env(var, int, default, strict=strict, split=split, catch=(ValueError,))


@overload
def read_float(var: str, default: float = 0.0, *, strict: bool = True, split: str) -> list[float]: ...
@overload
def read_float(var: str, default: float = 0.0, *, strict: bool = True, split: None = None) -> float: ...
def read_float(var: str, default: float = 0.0, *, strict: bool = True, split: str | None = None) -> float | list[float]:
    """Read ``float`` variable.

    Args:
        var: Variable to read.
        default: Default value to use when `var` is not set (or blank).
        strict: If ``False``, always fall back to `default` instead of raising.
        split: Character to split on. Returns ``list[float]`` when set.

    Returns:
        An ``float`` value, or a list thereof (if `split` is set).

    Examples:
        Basic usage.

        >>> import os
        >>> os.environ["MY_FLOAT"] = "1.0"
        >>> read_float("MY_FLOAT")
        1.0

        When using ``strict=False``, unmapped values are converted to the `default` instead of raising.

        >>> os.environ["MY_FLOAT"] = "not-a-float"
        >>> read_float("MY_FLOAT", strict=False)
        0.0

        When using `split`, elements are cleaned individually and blank items are skipped.

        >>> os.environ["MY_FLOAT_LIST"] = "0, 1.1, , 5"
        >>> read_float("MY_FLOAT_LIST", split=",")
        [0.0, 1.1, 5.0]

        Conversion must succeed for all elements, or the `default` will be returned.
    """
    return read_env(var, float, default, strict=strict, split=split, catch=(ValueError,))
